

[string] $ResultAppName = 'Corinth.Corinth3D'
[string] $ResultAppFamilyName = 'Corinth.Corinth3D_e4zw64d4dn61a'

#region WINDOWS STATUS AND CONFIGURATION
# - windows features
<#
[string]function GetAppInstallerVersion
{
    $AppInstallerVersion = (Get-AppxPackage -Name Microsoft.DesktopAppInstaller).Version
    return $AppInstallerVersion
}
#>

function Test-Sideloading
{
    [CmdletBinding()]
    $SideloadingEnabled = $False
    $SideloadingSettingsPath = 'Registry::HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\AppModelUnlock'
    if (Test-Path $SideloadingSettingsPath)
    {
        $SideloadingEnabled = (Get-ItemProperty -LiteralPath $SideloadingSettingsPath).AllowAllTrustedApps -eq 1
    }
    return $SideloadingEnabled
}

function Test-AppDataSharing
{
    [CmdletBinding()]
    $SharingEnabled = $False
    $SharingSettingsPath = 'Registry::HKEY_LOCAL_MACHINE\SOFTWARE\Policies\Microsoft\Windows\CurrentVersion\AppModel\StateManager'
    if (Test-Path $SharingSettingsPath)
    {
        $SharingEnabled = (Get-ItemProperty -LiteralPath $SharingSettingsPath).AllowSharedLocalAppData -eq 1
    }
    return $SharingEnabled
}

function Enable-Sideloading
{
    [CmdletBinding]
    $SideloadingSettingsPath = 'Registry::HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\AppModelUnlock'
    if (-not (Test-Path $SideloadingSettingsPath))
    {
        New-Item -Path $SideloadingSettingsPath -Force | Out-Null
    }
    New-ItemProperty -Path $SideloadingSettingsPath -Name 'AllowAllTrustedApps' -Value 1 -PropertyType DWORD -Force | Out-Null
    # alternativelly: REG ADD 'HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\AppModelUnlock' /t REG_DWORD /f /v 'AllowAllTrustedApps' /d '1'
}

function Enable-AppDataSharing
{
    [CmdletBinding]
    $SharingSettingsPath = 'Registry::HKEY_LOCAL_MACHINE\SOFTWARE\Policies\Microsoft\Windows\CurrentVersion\AppModel\StateManager'
    if (-not (Test-Path $SharingSettingsPath))
    {
        New-Item -Path $SharingSettingsPath -Force | Out-Null
    }
    New-ItemProperty -Path $SharingSettingsPath -Name 'AllowSharedLocalAppData' -Value 1 -PropertyType DWORD -Force | Out-Null
    # alternativelly: REG ADD 'HKEY_LOCAL_MACHINE\SOFTWARE\Policies\Microsoft\Windows\CurrentVersion\AppModel\StateManager' /t REG_DWORD /f /v 'AllowSharedLocalAppData' /d '1'
}

# - script execution privileges
function Test-AdminPrivileges
{
    return ([Security.Principal.WindowsPrincipal][Security.Principal.WindowsIdentity]::GetCurrent()).IsInRole([Security.Principal.WindowsBuiltInRole]::Administrator)
}

function Test-SystemPrivileges
{
    return ([Security.Principal.WindowsIdentity]::GetCurrent()).IsSystem
}
#endregion

#region CERTIFICATE INFO
# - install (only for debug)
function Install-Certificate
{
    [CmdletBinding()]
    param (
        [Parameter(Mandatory)]
        [string]$CertificatePath
    )
    certutil.exe -addstore TrustedPeople $CertificatePath | Out-Null
    #certutil.exe -addstore TrustedPublishers $CertificatePath | Out-Null
    return ($LastExitCode -ge 0) # True if successful
}

# - check package
function Test-PackageSignature
{
    [CmdletBinding()]
    param (
        [Parameter(Mandatory)]
        [string]$PackagePath
    )
    $PackageSignature = Get-AuthenticodeSignature $PackagePath -InformationAction Continue -ErrorAction SilentlyContinue
    return ($PackageSignature.Status -eq "Valid")
}
#endregion

#region APP DATA PATHS
function Get-InstallPath
{
    [CmdletBinding()]
    param (
        [Microsoft.Windows.Appx.PackageManager.Commands.AppxPackage]$PackageInfo = $null # type: Microsoft.Dism.Commands.AppxPackageObject or Microsoft.Windows.Appx.PackageManager.Commands.AppxPackage ?
    )
    if (-not $PackageInfo)
    {
        $PackageInfo = (Get-InstalledAppInfo -AppName $ResultAppName -AllUsers $True)
    }
    if ($PackageInfo)
    {
        return $PackageInfo.InstallLocation # $null if not installed
    }
    else
    {
        return $null
    }
}

function Get-SharedDataPath
{
    [CmdletBinding()]
    param (
        [string]$AppFamilyName = $ResultAppFamilyName
    )
    $SharedPath = "$Env:ProgramData\Microsoft\Windows\AppRepository\Families\$AppFamilyName\SharedLocal\"
    if (Test-Path $SharedPath) { return $SharedPath }
    else { return $null }
}

function Get-LocalDataPath
{
    [CmdletBinding()]
    param (
        [string]$AppFamilyName = $ResultAppFamilyName
    )
    $LocalPath = "$Env:LocalAppData\Packages\$AppFamilyName\LocalState\"
    if (Test-Path $LocalPath) { return $LocalPath }
    else { return $null }
}
#endregion

#region INSTALLED APP INFO
function Get-ProvisionedAppInfo
{
    [CmdletBinding()]
    param (
        [string]$AppFamilyName = $ResultAppFamilyName
    )
    $AppFNSplit = $AppFamilyName.Split('_')
    $AppName = $AppFNSplit[0]
    $PublisherID = $AppFNSplit[-1]
    return (Get-AppxProvisionedPackage -Online | Where-Object { $_.PackageName -like "$AppName*$PublisherID" })
}

function Get-InstalledAppInfo
{
    [CmdletBinding()]
    param (
        [string]$AppName = $ResultAppName,
        [bool]$AllUsers = $False
    )
    if (-not $AllUsers) # installed for current user
    {
        return (Get-AppxPackage -Name $AppName)
    }
    else # installed for any user (needs admin privileges)
    {
        return (Get-AppxPackage -Name $AppName -AllUsers)    
    }
}

function Test-AppWasActivated
{
    [CmdletBinding()]
    param (
        [string]$AppFamilyName = $ResultAppFamilyName
    )
    # checks whether Corinth license response exists
    # this doesn't mean the license is active (content unlocked) now, only that it had been activated at some point in the past - the license could be expired
    $AppDataPath = Get-SharedDataPath $AppFamilyName # shared local path - activated for all users
    if ($AppDataPath)
    {
        $LicenseResponsePath = Join-Path $AppDataPath 'Licenses\'
        if ((Test-Path $LicenseResponsePath) -and (Get-ChildItem (Join-Path $LicenseResponsePath '*.plic')).Count -gt 0)
        {
            return $True
        }
    }

    $AppDataPath = Get-LocalDataPath $AppFamilyName # local path - activated for current user
    if ($AppDataPath)
    {
        $LicenseResponsePath = Join-Path $AppDataPath 'Licenses\'
        if ((Test-Path $LicenseResponsePath) -and (Get-ChildItem (Join-Path $LicenseResponsePath '*.plic')).Count -gt 0)
        {
            return $True
        }
    }
    return $False
}
#endregion

#region APP INSTALLATION/REGISTRATION
function Install-AppxPackage
{
    [CmdletBinding()]
    param (
        [Parameter(Mandatory)]
        [string]$PackagePath,
        [string[]]$DependencyPackagePaths,
        [bool]$AllUsers=$False
    )

    if ($AllUsers) # provisioning (for current and "new" users)
    {
        if ($DependencyPackagePaths.Count -gt 0)
        {
            Add-AppxProvisionedPackage -Online -PackagePath $PackagePath -DependencyPackagePath $DependencyPackagePaths -SkipLicense
            return $?
        }
        else
        {
            Add-AppxProvisionedPackage -Online -PackagePath $PackagePath -SkipLicense
            return $?
        }
        #alternativelly: dism.exe /Online /add-ProvisionedAppxPackage /PackagePath $PackagePath /DependencyPackagePath $DependencyPackagePaths /SkipLicense # same as Add-AppxProvisionedPackage; needs a path to DISM
    }
    else # addition (for the current user only)
    {
        if ($DependencyPackagePaths.Count -gt 0)
        {
            Add-AppxPackage -Path $PackagePath -DependencyPath $DependencyPackagePaths -ForceApplicationShutdown
            return $?
        }
        else
        {
            Add-AppxPackage -Path $PackagePath -ForceApplicationShutdown
            return $?
        }
    }
}

function Register-AppxPackage
{
    [CmdletBinding()]
    param (
        [string]$AppFamilyName = $ResultAppFamilyName
    )

    Add-AppxPackage -RegisterByFamilyName -MainPackage $AppFamilyName
    return $?
}
#endregion

Export-ModuleMember -Function *
Export-ModuleMember -Variable $ResultAppFamilyName

# SIG # Begin signature block
# MIIfHQYJKoZIhvcNAQcCoIIfDjCCHwoCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUD2D9i9AqdbIX8V1oKN5iKpOw
# DlSgghkEMIIFDDCCA/SgAwIBAgIRAJ0ixooo0boZelXoKg5DAFQwDQYJKoZIhvcN
# AQELBQAwfDELMAkGA1UEBhMCR0IxGzAZBgNVBAgTEkdyZWF0ZXIgTWFuY2hlc3Rl
# cjEQMA4GA1UEBxMHU2FsZm9yZDEYMBYGA1UEChMPU2VjdGlnbyBMaW1pdGVkMSQw
# IgYDVQQDExtTZWN0aWdvIFJTQSBDb2RlIFNpZ25pbmcgQ0EwHhcNMjEwNDA2MDAw
# MDAwWhcNMjIwMzI0MjM1OTU5WjBPMQswCQYDVQQGEwJDWjEOMAwGA1UEBwwFUHJh
# aGExFzAVBgNVBAoMDkNvcmludGggcy5yLm8uMRcwFQYDVQQDDA5Db3JpbnRoIHMu
# ci5vLjCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBANCSI5UhlyFsK0/1
# OsfO3rajcKGxMp/Qo/uUs74BI5IaZ+kSRDgvzI8Ay9/O/x3EiO6gW1jDMbpOhHHR
# x72DmDavSP4JTvkTaPvwhc7eoCGCsyhi8J1Kca4wtrpUushcfaBn7CS9RVX6rDIB
# uTPsyVuPaMCjj/f1AWB+tE+4Xd1LVSPa+rs7akopoMmtrZ9jA7bDavclgAUY83jg
# BvRhqSmawCglc4o/Ap21iPvaUDfDGwPTz5ff6O/sKc+dFmpHH6RSD/zIy5vJ/2lm
# 5t2o11Oe8f3HshraKbemmLXHmFeoBJuySzWmlrCD/Q4jDRlkAeCa0y6Ttcu1Q81g
# Md/60o0CAwEAAaOCAbQwggGwMB8GA1UdIwQYMBaAFA7hOqhTOjHVir7Bu61nGgOF
# rTQOMB0GA1UdDgQWBBR9VxGgGc7reAaWPimwyFWfXcH3jzAOBgNVHQ8BAf8EBAMC
# B4AwDAYDVR0TAQH/BAIwADATBgNVHSUEDDAKBggrBgEFBQcDAzARBglghkgBhvhC
# AQEEBAMCBBAwSgYDVR0gBEMwQTA1BgwrBgEEAbIxAQIBAwIwJTAjBggrBgEFBQcC
# ARYXaHR0cHM6Ly9zZWN0aWdvLmNvbS9DUFMwCAYGZ4EMAQQBMEMGA1UdHwQ8MDow
# OKA2oDSGMmh0dHA6Ly9jcmwuc2VjdGlnby5jb20vU2VjdGlnb1JTQUNvZGVTaWdu
# aW5nQ0EuY3JsMHMGCCsGAQUFBwEBBGcwZTA+BggrBgEFBQcwAoYyaHR0cDovL2Ny
# dC5zZWN0aWdvLmNvbS9TZWN0aWdvUlNBQ29kZVNpZ25pbmdDQS5jcnQwIwYIKwYB
# BQUHMAGGF2h0dHA6Ly9vY3NwLnNlY3RpZ28uY29tMCIGA1UdEQQbMBmBF2RldmVs
# b3BlckBjb3JpbnRoM2QuY29tMA0GCSqGSIb3DQEBCwUAA4IBAQBPF0WmYhSqERFn
# q4B2EwlPbItCuQxOUBdVD8Z485t66uDfxs2wpJaHIfUdx/6sQahqt9vtwqNdmaIH
# JCV4cwfgQrEbxo7kuKt+Q11ZCsEqbCNsmfT4T2s3JBFCF8QpW/VfZP+DGT2mF+TB
# TG/0+/ZERTDXWVm26CBWTMTnUXkROV9Hwdym48hPx5WnkngqSa+byYJAS++dbUxf
# fl3+kYE2JHHLxguqbFQVbEluM3urkv4uAraoQCTB15Di1NuHeMkgVpDn6Q3nRdEf
# 0sHDLc/I1LPOuxwgGAW3gPkik/W8nsbl0a3n2d4EYRKaB9tJNlmI5c5XiE8BVnAb
# 8UdevLa3MIIF9TCCA92gAwIBAgIQHaJIMG+bJhjQguCWfTPTajANBgkqhkiG9w0B
# AQwFADCBiDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCk5ldyBKZXJzZXkxFDASBgNV
# BAcTC0plcnNleSBDaXR5MR4wHAYDVQQKExVUaGUgVVNFUlRSVVNUIE5ldHdvcmsx
# LjAsBgNVBAMTJVVTRVJUcnVzdCBSU0EgQ2VydGlmaWNhdGlvbiBBdXRob3JpdHkw
# HhcNMTgxMTAyMDAwMDAwWhcNMzAxMjMxMjM1OTU5WjB8MQswCQYDVQQGEwJHQjEb
# MBkGA1UECBMSR3JlYXRlciBNYW5jaGVzdGVyMRAwDgYDVQQHEwdTYWxmb3JkMRgw
# FgYDVQQKEw9TZWN0aWdvIExpbWl0ZWQxJDAiBgNVBAMTG1NlY3RpZ28gUlNBIENv
# ZGUgU2lnbmluZyBDQTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAIYi
# jTKFehifSfCWL2MIHi3cfJ8Uz+MmtiVmKUCGVEZ0MWLFEO2yhyemmcuVMMBW9aR1
# xqkOUGKlUZEQauBLYq798PgYrKf/7i4zIPoMGYmobHutAMNhodxpZW0fbieW15dR
# hqb0J+V8aouVHltg1X7XFpKcAC9o95ftanK+ODtj3o+/bkxBXRIgCFnoOc2P0tbP
# BrRXBbZOoT5Xax+YvMRi1hsLjcdmG0qfnYHEckC14l/vC0X/o84Xpi1VsLewvFRq
# nbyNVlPG8Lp5UEks9wO5/i9lNfIi6iwHr0bZ+UYc3Ix8cSjz/qfGFN1VkW6KEQ3f
# BiSVfQ+noXw62oY1YdMCAwEAAaOCAWQwggFgMB8GA1UdIwQYMBaAFFN5v1qqK0rP
# VIDh2JvAnfKyA2bLMB0GA1UdDgQWBBQO4TqoUzox1Yq+wbutZxoDha00DjAOBgNV
# HQ8BAf8EBAMCAYYwEgYDVR0TAQH/BAgwBgEB/wIBADAdBgNVHSUEFjAUBggrBgEF
# BQcDAwYIKwYBBQUHAwgwEQYDVR0gBAowCDAGBgRVHSAAMFAGA1UdHwRJMEcwRaBD
# oEGGP2h0dHA6Ly9jcmwudXNlcnRydXN0LmNvbS9VU0VSVHJ1c3RSU0FDZXJ0aWZp
# Y2F0aW9uQXV0aG9yaXR5LmNybDB2BggrBgEFBQcBAQRqMGgwPwYIKwYBBQUHMAKG
# M2h0dHA6Ly9jcnQudXNlcnRydXN0LmNvbS9VU0VSVHJ1c3RSU0FBZGRUcnVzdENB
# LmNydDAlBggrBgEFBQcwAYYZaHR0cDovL29jc3AudXNlcnRydXN0LmNvbTANBgkq
# hkiG9w0BAQwFAAOCAgEATWNQ7Uc0SmGk295qKoyb8QAAHh1iezrXMsL2s+Bjs/th
# AIiaG20QBwRPvrjqiXgi6w9G7PNGXkBGiRL0C3danCpBOvzW9Ovn9xWVM8Ohgyi3
# 3i/klPeFM4MtSkBIv5rCT0qxjyT0s4E307dksKYjalloUkJf/wTr4XRleQj1qZPe
# a3FAmZa6ePG5yOLDCBaxq2NayBWAbXReSnV+pbjDbLXP30p5h1zHQE1jNfYw08+1
# Cg4LBH+gS667o6XQhACTPlNdNKUANWlsvp8gJRANGftQkGG+OY96jk32nw4e/gdR
# EmaDJhlIlc5KycF/8zoFm/lv34h/wCOe0h5DekUxwZxNqfBZslkZ6GqNKQQCd3xL
# S81wvjqyVVp4Pry7bwMQJXcVNIr5NsxDkuS6T/FikyglVyn7URnHoSVAaoRXxrKd
# sbwcCtp8Z359LukoTBh+xHsxQXGaSynsCz1XUNLK3f2eBVHlRHjdAd6xdZgNVCT9
# 8E7j4viDvXK6yz067vBeF5Jobchh+abxKgoLpbn0nu6YMgWFnuv5gynTxix9vTp3
# Los3QqBqgu07SqqUEKThDfgXxbZaeTMYkuO1dfih6Y4KJR7kHvGfWocj/5+kUZ77
# OYARzdu1xKeogG/lU9Tg46LC0lsa+jImLWpXcBw8pFguo/NbSwfcMlnzh6cabVgw
# ggbsMIIE1KADAgECAhAwD2+s3WaYdHypRjaneC25MA0GCSqGSIb3DQEBDAUAMIGI
# MQswCQYDVQQGEwJVUzETMBEGA1UECBMKTmV3IEplcnNleTEUMBIGA1UEBxMLSmVy
# c2V5IENpdHkxHjAcBgNVBAoTFVRoZSBVU0VSVFJVU1QgTmV0d29yazEuMCwGA1UE
# AxMlVVNFUlRydXN0IFJTQSBDZXJ0aWZpY2F0aW9uIEF1dGhvcml0eTAeFw0xOTA1
# MDIwMDAwMDBaFw0zODAxMTgyMzU5NTlaMH0xCzAJBgNVBAYTAkdCMRswGQYDVQQI
# ExJHcmVhdGVyIE1hbmNoZXN0ZXIxEDAOBgNVBAcTB1NhbGZvcmQxGDAWBgNVBAoT
# D1NlY3RpZ28gTGltaXRlZDElMCMGA1UEAxMcU2VjdGlnbyBSU0EgVGltZSBTdGFt
# cGluZyBDQTCCAiIwDQYJKoZIhvcNAQEBBQADggIPADCCAgoCggIBAMgbAa/ZLH6I
# mX0BmD8gkL2cgCFUk7nPoD5T77NawHbWGgSlzkeDtevEzEk0y/NFZbn5p2QWJgn7
# 1TJSeS7JY8ITm7aGPwEFkmZvIavVcRB5h/RGKs3EWsnb111JTXJWD9zJ41OYOioe
# /M5YSdO/8zm7uaQjQqzQFcN/nqJc1zjxFrJw06PE37PFcqwuCnf8DZRSt/wflXMk
# PQEovA8NT7ORAY5unSd1VdEXOzQhe5cBlK9/gM/REQpXhMl/VuC9RpyCvpSdv7Qg
# sGB+uE31DT/b0OqFjIpWcdEtlEzIjDzTFKKcvSb/01Mgx2Bpm1gKVPQF5/0xrPnI
# hRfHuCkZpCkvRuPd25Ffnz82Pg4wZytGtzWvlr7aTGDMqLufDRTUGMQwmHSCIc9i
# VrUhcxIe/arKCFiHd6QV6xlV/9A5VC0m7kUaOm/N14Tw1/AoxU9kgwLU++Le8bwC
# KPRt2ieKBtKWh97oaw7wW33pdmmTIBxKlyx3GSuTlZicl57rjsF4VsZEJd8GEpoG
# LZ8DXv2DolNnyrH6jaFkyYiSWcuoRsDJ8qb/fVfbEnb6ikEk1Bv8cqUUotStQxyk
# SYtBORQDHin6G6UirqXDTYLQjdprt9v3GEBXc/Bxo/tKfUU2wfeNgvq5yQ1TgH36
# tjlYMu9vGFCJ10+dM70atZ2h3pVBeqeDAgMBAAGjggFaMIIBVjAfBgNVHSMEGDAW
# gBRTeb9aqitKz1SA4dibwJ3ysgNmyzAdBgNVHQ4EFgQUGqH4YRkgD8NBd0UojtE1
# XwYSBFUwDgYDVR0PAQH/BAQDAgGGMBIGA1UdEwEB/wQIMAYBAf8CAQAwEwYDVR0l
# BAwwCgYIKwYBBQUHAwgwEQYDVR0gBAowCDAGBgRVHSAAMFAGA1UdHwRJMEcwRaBD
# oEGGP2h0dHA6Ly9jcmwudXNlcnRydXN0LmNvbS9VU0VSVHJ1c3RSU0FDZXJ0aWZp
# Y2F0aW9uQXV0aG9yaXR5LmNybDB2BggrBgEFBQcBAQRqMGgwPwYIKwYBBQUHMAKG
# M2h0dHA6Ly9jcnQudXNlcnRydXN0LmNvbS9VU0VSVHJ1c3RSU0FBZGRUcnVzdENB
# LmNydDAlBggrBgEFBQcwAYYZaHR0cDovL29jc3AudXNlcnRydXN0LmNvbTANBgkq
# hkiG9w0BAQwFAAOCAgEAbVSBpTNdFuG1U4GRdd8DejILLSWEEbKw2yp9KgX1vDsn
# 9FqguUlZkClsYcu1UNviffmfAO9Aw63T4uRW+VhBz/FC5RB9/7B0H4/GXAn5M17q
# oBwmWFzztBEP1dXD4rzVWHi/SHbhRGdtj7BDEA+N5Pk4Yr8TAcWFo0zFzLJTMJWk
# 1vSWVgi4zVx/AZa+clJqO0I3fBZ4OZOTlJux3LJtQW1nzclvkD1/RXLBGyPWwlWE
# ZuSzxWYG9vPWS16toytCiiGS/qhvWiVwYoFzY16gu9jc10rTPa+DBjgSHSSHLeT8
# AtY+dwS8BDa153fLnC6NIxi5o8JHHfBd1qFzVwVomqfJN2Udvuq82EKDQwWli6YJ
# /9GhlKZOqj0J9QVst9JkWtgqIsJLnfE5XkzeSD2bNJaaCV+O/fexUpHOP4n2HKG1
# qXUfcb9bQ11lPVCBbqvw0NP8srMftpmWJvQ8eYtcZMzN7iea5aDADHKHwW5NWtMe
# 6vBE5jJvHOsXTpTDeGUgOw9Bqh/poUGd/rG4oGUqNODeqPk85sEwu8CgYyz8XBYA
# qNDEf+oRnR4GxqZtMl20OAkrSQeq/eww2vGnL8+3/frQo4TZJ577AWZ3uVYQ4SBu
# xq6x+ba6yDVdM3aO8XwgDCp3rrWiAoa6Ke60WgCxjKvj+QrJVF3UuWp0nr1Irpgw
# ggcHMIIE76ADAgECAhEAjHegAI/00bDGPZ86SIONazANBgkqhkiG9w0BAQwFADB9
# MQswCQYDVQQGEwJHQjEbMBkGA1UECBMSR3JlYXRlciBNYW5jaGVzdGVyMRAwDgYD
# VQQHEwdTYWxmb3JkMRgwFgYDVQQKEw9TZWN0aWdvIExpbWl0ZWQxJTAjBgNVBAMT
# HFNlY3RpZ28gUlNBIFRpbWUgU3RhbXBpbmcgQ0EwHhcNMjAxMDIzMDAwMDAwWhcN
# MzIwMTIyMjM1OTU5WjCBhDELMAkGA1UEBhMCR0IxGzAZBgNVBAgTEkdyZWF0ZXIg
# TWFuY2hlc3RlcjEQMA4GA1UEBxMHU2FsZm9yZDEYMBYGA1UEChMPU2VjdGlnbyBM
# aW1pdGVkMSwwKgYDVQQDDCNTZWN0aWdvIFJTQSBUaW1lIFN0YW1waW5nIFNpZ25l
# ciAjMjCCAiIwDQYJKoZIhvcNAQEBBQADggIPADCCAgoCggIBAJGHSyyLwfEeoJ7T
# B8YBylKwvnl5XQlmBi0vNX27wPsn2kJqWRslTOrvQNaafjLIaoF9tFw+VhCBNToi
# Noz7+CAph6x00BtivD9khwJf78WA7wYc3F5Ok4e4mt5MB06FzHDFDXvsw9njl+nL
# GdtWRWzuSyBsyT5s/fCb8Sj4kZmq/FrBmoIgOrfv59a4JUnCORuHgTnLw7c6zZ9Q
# BB8amaSAAk0dBahV021SgIPmbkilX8GJWGCK7/GszYdjGI50y4SHQWljgbz2H6p8
# 18FBzq2rdosggNQtlQeNx/ULFx6a5daZaVHHTqadKW/neZMNMmNTrszGKYogwWDG
# 8gIsxPnIIt/5J4Khg1HCvMmCGiGEspe81K9EHJaCIpUqhVSu8f0+SXR0/I6uP6Vy
# 9MNaAapQpYt2lRtm6+/a35Qu2RrrTCd9TAX3+CNdxFfIJgV6/IEjX1QJOCpi1arK
# 3+3PU6sf9kSc1ZlZxVZkW/eOUg9m/Jg/RAYTZG7p4RVgUKWx7M+46MkLvsWE990K
# ndq8KWw9Vu2/eGe2W8heFBy5r4Qtd6L3OZU3b05/HMY8BNYxxX7vPehRfnGtJHQb
# LNz5fKrvwnZJaGLVi/UD3759jg82dUZbk3bEg+6CviyuNxLxvFbD5K1Dw7dmll6U
# Mvqg9quJUPrOoPMIgRrRRKfM97gxAgMBAAGjggF4MIIBdDAfBgNVHSMEGDAWgBQa
# ofhhGSAPw0F3RSiO0TVfBhIEVTAdBgNVHQ4EFgQUaXU3e7udNUJOv1fTmtufAdGu
# 3tAwDgYDVR0PAQH/BAQDAgbAMAwGA1UdEwEB/wQCMAAwFgYDVR0lAQH/BAwwCgYI
# KwYBBQUHAwgwQAYDVR0gBDkwNzA1BgwrBgEEAbIxAQIBAwgwJTAjBggrBgEFBQcC
# ARYXaHR0cHM6Ly9zZWN0aWdvLmNvbS9DUFMwRAYDVR0fBD0wOzA5oDegNYYzaHR0
# cDovL2NybC5zZWN0aWdvLmNvbS9TZWN0aWdvUlNBVGltZVN0YW1waW5nQ0EuY3Js
# MHQGCCsGAQUFBwEBBGgwZjA/BggrBgEFBQcwAoYzaHR0cDovL2NydC5zZWN0aWdv
# LmNvbS9TZWN0aWdvUlNBVGltZVN0YW1waW5nQ0EuY3J0MCMGCCsGAQUFBzABhhdo
# dHRwOi8vb2NzcC5zZWN0aWdvLmNvbTANBgkqhkiG9w0BAQwFAAOCAgEASgN4kEIz
# 7Hsagwk2M5hVu51ABjBrRWrxlA4ZUP9bJV474TnEW7rplZA3N73f+2Ts5YK3lcxX
# VXBLTvSoh90ihaZXu7ghJ9SgKjGUigchnoq9pxr1AhXLRFCZjOw+ugN3poICkMIu
# k6m+ITR1Y7ngLQ/PATfLjaL6uFqarqF6nhOTGVWPCZAu3+qIFxbradbhJb1FCJeA
# 11QgKE/Ke7OzpdIAsGA0ZcTjxcOl5LqFqnpp23WkPnlomjaLQ6421GFyPA6FYg2g
# XnDbZC8Bx8GhxySUo7I8brJeotD6qNG4JRwW5sDVf2gaxGUpNSotiLzqrnTWgufA
# iLjhT3jwXMrAQFzCn9UyHCzaPKw29wZSmqNAMBewKRaZyaq3iEn36AslM7U/ba+f
# XwpW3xKxw+7OkXfoIBPpXCTH6kQLSuYThBxN6w21uIagMKeLoZ+0LMzAFiPJkeVC
# A0uAzuRN5ioBPsBehaAkoRdA1dvb55gQpPHqGRuAVPpHieiYgal1wA7f0GiUeaGg
# no62t0Jmy9nZay9N2N4+Mh4g5OycTUKNncczmYI3RNQmKSZAjngvue76L/Hxj/5Q
# uHjdFJbeHA5wsCqFarFsaOkq5BArbiH903ydN+QqBtbD8ddo408HeYEIE/6yZF7p
# sTzm0Hgjsgks4iZivzupl1HMx0QygbKvz98xggWDMIIFfwIBATCBkTB8MQswCQYD
# VQQGEwJHQjEbMBkGA1UECBMSR3JlYXRlciBNYW5jaGVzdGVyMRAwDgYDVQQHEwdT
# YWxmb3JkMRgwFgYDVQQKEw9TZWN0aWdvIExpbWl0ZWQxJDAiBgNVBAMTG1NlY3Rp
# Z28gUlNBIENvZGUgU2lnbmluZyBDQQIRAJ0ixooo0boZelXoKg5DAFQwCQYFKw4D
# AhoFAKB4MBgGCisGAQQBgjcCAQwxCjAIoAKAAKECgAAwGQYJKoZIhvcNAQkDMQwG
# CisGAQQBgjcCAQQwHAYKKwYBBAGCNwIBCzEOMAwGCisGAQQBgjcCARUwIwYJKoZI
# hvcNAQkEMRYEFJ2ubGREw2B5NBo6AtyFZVStRsNoMA0GCSqGSIb3DQEBAQUABIIB
# AJoJETj3qAfonQ9GREgQLQGL9VffWAJwHK+mdLhXJq0/DZY8aE1llckG5rHGx2ih
# GIAmZYHhb86PRwdwGJLgi3Si0GkR5v6QLB9//Sc8Ye2zUh1JsXtqQDxkP41sEj2U
# hQqSP413XVjPakda4WG2u2lCZSjlRwft7JOFeGhL8yILx8aY0DEDkk03UdALBtb2
# 20l0poudsMFqn/VOFWNGSTyYROk6NWTy0A8GkHbDTSp+QphbFpbp0aZe3bKuCFup
# zE5XgJ4xYIll3oIsSqHP7WU/ehq0+2nVUHop2fcW6g+xgfd+pciQpOA8OLseRLX0
# XUeRofBcaOyLyQ8ZOYCLSpehggNMMIIDSAYJKoZIhvcNAQkGMYIDOTCCAzUCAQEw
# gZIwfTELMAkGA1UEBhMCR0IxGzAZBgNVBAgTEkdyZWF0ZXIgTWFuY2hlc3RlcjEQ
# MA4GA1UEBxMHU2FsZm9yZDEYMBYGA1UEChMPU2VjdGlnbyBMaW1pdGVkMSUwIwYD
# VQQDExxTZWN0aWdvIFJTQSBUaW1lIFN0YW1waW5nIENBAhEAjHegAI/00bDGPZ86
# SIONazANBglghkgBZQMEAgIFAKB5MBgGCSqGSIb3DQEJAzELBgkqhkiG9w0BBwEw
# HAYJKoZIhvcNAQkFMQ8XDTIxMDQxMzE2Mzk1MlowPwYJKoZIhvcNAQkEMTIEMOUW
# RKcSiVuAg/RYFBmdwz+hvru1GxNsjTCEH6qi3iCZmAR4f69xpIMwg6FSbDIkMDAN
# BgkqhkiG9w0BAQEFAASCAgA9VVDuM/laEpkDi1I9UnAdj0EdVevO0ygFKwjiUr5z
# SbHRv4uw6rb5k/KgDLohamE9PQV6gjNwEgwoYrIIg6UaZKtUrlUMSipgy7YUXkq5
# CFT5D7F3qmeSALTwPGr+cov9Ev5jnf89/FjSRgL9VKIAHlwAjnOpGRoQOBWTUIvu
# Fl9QlGKl88D3KmFgAQOxEaPH/upY2d1q201r9Cxr+GdqXj239Q3r5JxoJIb95Jik
# wvL+pnNFvzimo9H53fDSu29FQJYNGgRFpviFsbsQ4VNDmyUQWTCpToanZWeZ08Rs
# HDqyZwq3Fi5suvrHZLJYwgjTnRLb7tGBZOncUSbe9Iu8Fu2nL5jShPLS4AjqGJMY
# wKqHooQ+bMMN3TWeraOSaGVqxeUSOY6Dwul1qtryiblQwFy06dfs6GVF+wznu2dw
# O1KrdVG0Helti2ZkNDGGTH3dQneR2cveo9/7SeBC1qMwxS9efctP66GxU+yC/oAT
# hHTqoRmrpIdgkNr9cXaWKYNyteMZ4qXw60U3c6E7is7tDWPTKZyiuq+YYyRf5P3D
# LUKGqUIQSICprUktQ6AqVrmv+4F/NYphUniyF4EPNodqXJ/m6OJ9o8EVnHvmrPIR
# CvYKbwD4NbS8boQK0PtLWql33GFfy9jXkypqfEp+JDanEPXxsGh2K7+Guo5kQxzo
# Rg==
# SIG # End signature block
