
#
# Corinth Installation Script.
#
# Disclaimer:
#	The goal of this script is to demonstrate the possibilities of Corinth app deployment automation and to provide inspiration for creating your own solution.
#	This script will try and run itself with administrative or system privileges and may change various computer settings, including registry values, installed apps and trusted certificates.
#	Examine the code carefully before running it and make sure you understand its implications. Run this script at your own risk. No warranty provided.
#
# What this script does:
#	- logs the current state/configuration of the machine
#	- relaunch itself to try to get administrative privileges (self elevation)
#	- enable sideloading (via registry)
#	- enable app data sharing between users (via registry)
#	- add the Corinth app for current user (via staging and registering) or for all users on current machine (via provisioning)
#	- prepare app activation on the current machine (via license input file)
#	- sets an update policy/strategy (via xml)
#
# Recommended configuration:
#	- Windows 10
#		- architecture: 64-bit
#		- version: 1803 or higher
#		- edition: Pro or higher
#
# Parameters:
#	- string parameters:
#		- PackagePath: Path to the installation package. Any dependency packages are expected to be inside a folder named "Dependecies\[architecture]" (typically "Dependencies\x64") right next to the package file.
#			Default: The script will try to find a package within its folder (appx/msix/bundle right next to the script file) if no path specified.
#			Failure action: If there is no package available on the specified path or if no path was specified and there is no single package in the default folder, the installation fails.
#		- LicensePath: Path to a license file which should be used to activate the app and unlock its full content during its first launch.
#			Default: Will try to use a file named "license_input" located next to this script if no path specified.
#			Failure action: If there is no file available on the specified path or if no path was specified and there is no file on the default path, the script will fail. This bahaviour can be overridden by using the SkipLicense option (app will be installed but it won't be activated).
#		- CertificatePath: Path to a certificate which should be installed before installing the app. Mostly used for internal testing when the app is not yet signed by a publicly trusted certificate.
#			Default: No certificate will be installed if no path specified.
#			Failure action: If the certificate cannot be found, a warning will be issued and the script continues executing.
#		- UpdateStrategy: Species the preferred way of user interaction when handling updates. Allowed values: User, Admin or Force.
#				User: Users are supposed to handle the updates himself: upon update discovery, a dialog is shown with the options to install the update or skip.
#				Admin: A user is supposed to tell the admin that there is an update, and the admin should trigger the update (via app menu) at an appropriate time. This gives the admin more control over the state of the machine.
#				Force: The app updates itself automatically without any user interaction. This option does not demand any user attention, but it gives the user/admin very little control over the time of the updates.
#			Default: No settings file will be created, the app will use its default settings (User).
#			Failure action: If an invalid option is specified (i.e. other than the above-mentioned three), the script fails, just as with any other unrecognized parameter.
#	- switch parameters:
#		- InstallFromWeb: The script will download latest version of the installation packages from the internet and install them. Overrides PackagePath parameter.
#			Default: See PackagePath parameter.
#		- RegisterOnly: The script will only register (not install) the app for the current user. No further action will be taken, all other parameters (excluding Verbose, Help, PrintLog, PrintConfigReport and WhatIf) will be ignored. Presumes that the app is already installed on the computer.
#			Default: If the switch is not provided, the app will be installed and related steps will be performed.
#		- SingleUser: Installs the app only for the executing user.
#			Default: The app will be provisioned for all users on the computer if not otherwise specified.
#		- ForceInstall: Will try to re-install the app even if it is already installed.
#			Default: The installation (and all the related steps, such as certificate installation and sideloading activation) will be skipped when there is any version of the app already available for current user.
#		- SkipLicense: Will not provide any license input. The app can be activated later by this script or manually within the app by specifying the license code. The LicensePath patameter will be ignored.
#			Default: The script will try to copy a license input file (either from a location specified via parameter or from a default location next to this script) to the app data, so that it can be activated. It will throw an error, if the license input is unavailable.
#		- ForceLicense: Will copy the license input file to the app data even if the app had been previously activated. Overrides SkipLicense.
#			Default: If the script finds data sugesting that the app has been activated in the past, it will not try to provide the app with a license input file. This is to prevent the app from excessive number of attempt to activate itself - it tries to go through the activation process each time the license input changes.
#		- Verbose: Enables console output. Useful when running the script manually.
#			Default: Silent execution (except for the event logging and errors).
#		- Help: Prints help similar to this description and exits.
#		- PrintLog: Prints log and exits.
#		- PrintConfigReport: Prints configuration check report (i.e. Windows version, sideloading settings, installed version of the app etc.) and exits.
#		- WhatIf: Prints a summary of expected steps without actually executing them.
#		- SkipElevate: Doesn't relaunch to gain administrator privileges, just fails if it's not elevated. Used internally.
#
# Logging:
#	- this script logs its activity via Events that can be examined though the Event Viewer (default location: see $EventLogName)
#	- codes (EventId):
#		- first digit => type
#			- <1000 - pre-execution checks and reports, result summaries
#			- 1xxx - OK - standard operation, info, success
#			- 2xxx - OK - skipped steps (e.g. already valid settings; possibly due to repeated execution of the script)
#			- 3xxx - warnings
#			- 4xxx - errors
#		- middle digits ("hundreds") => script step
#			- x0xx - preparation (i.e. existing-installation check, parameter processing, event log creation, execution planning)
#			- x1xx - self-elevation (relaunch), script privileges check, launch info logging
#			- x2xx - installation packages acquisition (app + dependencies)
#			- x3xx - certificate check/installation
#			- x4xx - sideloading activation
#			- x5xx - data sharing activation
#			- x6xx - installation/provisioning
#			- x7xx - license input and update settings file placement
#			- x8xx - cleanup
#			- x9xx - execution summary
#		- last two digits => substep (e.g. checks before, during and after the step) and variant (e.g. too few vs. too many files specified)
#

param (
	[string]$PackagePath,
	[string]$CertificatePath, # only for development purposes
	[string]$LicensePath,
	[string]$UpdateStrategy,
	[switch]$InstallFromWeb = $False,
	[switch]$RegisterOnly = $False,
	[switch]$SingleUser = $False,
	[switch]$ForceInstall = $False,
	[switch]$SkipLicense = $False,
	[switch]$ForceLicense = $False,
	[switch]$Verbose = $False,
	[switch]$Help = $False,
	[switch]$PrintLog = $False,
	[switch]$PrintConfigReport = $False,
	[switch]$WhatIf = $False,
	[switch]$SkipElevate = $False, # only used internally by the script during relaunch
	[string]$ModulePath # only used internally by the script during relaunch
)

enum PackageSource
{
	Parameter
	Folder
	Download
}

enum UpdateStrategy
{
	User # user is allowed to install updates from the start
	Admin # user should only ask the admin, admin should then update the app
	Force # the app should just update itself without telling anyone
}

enum StepProgress
{
	NotExecuted
	Success
	Failure
}

Set-StrictMode -Version Latest # DEBUG

$ScriptVersion = '1.2.2203.0'
$PackageUrl = 'https://app-win-07.corinth3d.com/appx/Corinth3D_Sideload_7.186.278.0_Master_Test/Corinth3D_Sideload_7.186.278.0_x64_Master.msixbundle'
$DependencyUrls = @(
	'https://app-win-07.corinth3d.com/appx/Corinth3D_Sideload_7.186.278.0_Master_Test/Dependencies/x64/Microsoft.NET.Native.Framework.1.3.appx',
	'https://app-win-07.corinth3d.com/appx/Corinth3D_Sideload_7.186.278.0_Master_Test/Dependencies/x64/Microsoft.NET.Native.Runtime.1.4.appx',
	'https://app-win-07.corinth3d.com/appx/Corinth3D_Sideload_7.186.278.0_Master_Test/Dependencies/x64/Microsoft.VCLibs.x64.14.00.appx'
)

#region LOGGING SETUP
$EventLogName = 'Application'
$EventLogSource = 'Corinth Installation Script'

function CreateEventLog
{
	$HasEventSource = $False
	try
	{
		$HasEventSource = [System.Diagnostics.EventLog]::SourceExists($EventLogSource)
		if (-not $HasEventSource)
		{
			New-EventLog -LogName $EventLogName -Source $EventLogSource
		}
	}
	catch{}
}

function WriteEventLog($EntryType, $EventID, $EventMsg)
{
	try
	{
		if ([System.Diagnostics.EventLog]::SourceExists($EventLogSource))
		{
			Write-EventLog -LogName $EventLogName -Source $EventLogSource -EntryType $EntryType -EventId $EventID -Message $EventMsg
		}
	}
	catch{}
	finally
	{
		if ($EntryType -eq 'Error')
		{
			SendMixpanelError $EventID | Out-Null
			Write-Error "Error ($EventID): $EventMsg"
		}
		elseif ($Verbose)
		{
			Write-Host "$EntryType ($EventID): $EventMsg"
		}
	}
}
#endregion

if (-not $ModulePath) { $ModulePath = Join-Path (Split-Path -Parent (Get-Variable MyInvocation).Value.MyCommand.Path) 'CCInstallUtils.psd1' } # default is next to the script
Import-Module $ModulePath
if (-not $?)
{
	$Msg = 'Unable to load the Corinth installation PowerShell module (CCInstallUtils.psm1). Please provide the module and try again.'
	CreateEventLog
	WriteEventLog Error 9999 $Msg
	exit 9999
}

#region SCRIPT EXECUTION PREPARATION
# holds information about the script itself (version, module version, etc.)
class ScriptMetaInfo
{
	[string]$ScriptPath = $null
	[string]$ScriptDir = $null
	[string]$ScriptVersion = $ScriptVersion
	[string]$ScriptOrigin = $null
	[string]$ModuleVersion = $null
	[string]$ModuleOrigin = $null

	ScriptMetaInfo([object]$MyInvocation) # TODO: better type
	{
		# paths
		$this.ScriptPath = (Get-Variable MyInvocation).Value.MyCommand.Path
		$this.ScriptDir = Split-Path -Parent $this.ScriptPath
		# version
		$ScriptSignature = Get-AuthenticodeSignature $this.ScriptPath
		$this.ScriptOrigin = $this.GetSignatureDescription($ScriptSignature)
		$ModuleInfo = Get-Module CCInstallUtils
		$this.ModuleVersion = $ModuleInfo.Version
		$ModuleSignature = Get-AuthenticodeSignature $ModuleInfo.Path
		$this.ModuleOrigin = $this.GetSignatureDescription($ModuleSignature)
	}

	# -  signature check
	[string] GetSignatureDescription($Signature)
	{
		$CCCertThumbprint = '19D0D66516C733424581DA77556130F048B68044'
		if ($Signature.Status -ne 'NotSigned' -and $Signature.SignerCertificate)
		{
			if ($Signature.Status -eq 'HashMismatch') { return 'custom' } # modified
			elseif ($Signature.SignerCertificate.Thumbprint -eq $CCCertThumbprint) # original certificate
			{
				if ($Signature.Status -eq 'Valid') { return 'orig' } # trusted
				else { return 'debug' } # not trusted
			}
			else { return 'resigned' } # signed by a different certificate
		}
		else { return 'anon' } # unsigned
	}

	# - script identification (useful when trying to determine which version of the script was executed)
	[string] GetScriptID()
	{
		$ScriptID = 'Corinth installation script v' + $this.ScriptVersion + ' (' + $this.ScriptOrigin + ') '
		# module version
		$ScriptID += 'with CCInstallUtils module v' + $this.ModuleVersion + ' (' + $this.ModuleOrigin + ')'
		return $ScriptID
	}
}

# holds information about existing app installation (availability, version, activation)
class AppInstallationInfo
{
	[string]$AppInstalledVersionUser = 'Unknown'
	[string]$AppInstalledVersionAny = 'Unknown'
	[string]$AppProvisionedVersion = 'Unknown'
	[bool]$AppWasInstalled = $False # any installation found
	[bool]$AppWasActivated = $False

	AppInstallationInfo()
	{
		$this.AppWasInstalled = $False
		$UserInstalledAppInfo = Get-InstalledAppInfo
		if ($UserInstalledAppInfo)
		{
			$this.AppInstalledVersionUser = $UserInstalledAppInfo.Version
			$this.AppWasInstalled = $True
		}
		else { $this.AppInstalledVersionUser = 'None' }
		$AnyInstalledAppInfo = $null
		$ProvisionedAppInfo = $null
		if ((Test-AdminPrivileges) -or (Test-SystemPrivileges)) # needs higher privileges to access info about apps installed for other users
		{
			$AnyInstalledAppInfo = Get-InstalledAppInfo -AllUsers $True
			if ($AnyInstalledAppInfo)
			{
				$this.AppInstalledVersionAny = $AnyInstalledAppInfo.Version 
				$this.AppWasInstalled = $True
			}
			else { $this.AppInstalledVersionAny = 'None' }
			$ProvisionedAppInfo = Get-ProvisionedAppInfo
			if ($ProvisionedAppInfo) {
				$this.AppProvisionedVersion = $ProvisionedAppInfo.Version
				$this.AppWasInstalled = $True
			}
			else { $this.AppProvisionedVersion = 'None' }
		}
		else
		{
			$this.AppInstalledVersionAny = 'Unknown'
			$this.AppProvisionedVersion = 'Unknown'
		}
		if ($UserInstalledAppInfo -or $AnyInstalledAppInfo -or $ProvisionedAppInfo)
		{
			$this.AppWasActivated = Test-AppWasActivated
		}
	}
}

# determines which operations need to be executed (based on parameters and system settings) and tracks the progress (to evaluate the overall success of the script)
# implements the script main execution logic
class ScriptExecutionPlan
{
	# operations to be done
	[bool]$Elevate = $False
	[bool]$InstallCertificate = $False
	[bool]$EnableDataSharing = $False
	[bool]$EnableSideloading = $False
	[PackageSource]$PackageSource = [PackageSource]'Folder'
	[bool]$RegisterForSingleUser = $False # register package (must be already installed/staged)
	[bool]$InstallForSingleUser = $False # add package (stage and register)
	[bool]$InstallForAllUsers = $False # provision package
	[bool]$CopyLicense = $False
	[bool]$CreateUpdateStrategy = $False

	# extra (private) info
	hidden [bool]$Reinstall = $False # is already installed but should be installed (again) anyway
	hidden [bool]$FailElevate = $False # a combination of parameters that can't be satisfied given the current configuration

	# step result - setup upon step competion
	[StepProgress]$CertificateInstallProgress = [StepProgress]'NotExecuted'
	[StepProgress]$DataSharingSetupProgress = [StepProgress]'NotExecuted'
	[StepProgress]$SideloadingSetupProgress = [StepProgress]'NotExecuted'
	[StepProgress]$AppInstallationProgress = [StepProgress]'NotExecuted'
	[StepProgress]$LicenseInputSetupProgress = [StepProgress]'NotExecuted'
	[StepProgress]$UpdateStrategySetupProgress = [StepProgress]'NotExecuted'

	[hashtable]$LaunchParameters

	ScriptExecutionPlan([hashtable]$BoundParameters)
	{
		$this.LaunchParameters = $BoundParameters

		if ($BoundParameters['RegisterOnly'])
		{
			if (-not (Get-InstalledAppInfo)) { $this.RegisterForSingleUser = $True }
			return # no further action
		}

		# pick ops that really need to be done
		if (-not (Test-AppDataSharing)) { $this.EnableDataSharing = $True }

		if ($BoundParameters['ForceInstall'] -or ($this.NeedsInstall()))
		{
			# app not installed
			# -> enable sideloading (if needed)
			if (-not (Test-Sideloading)) { $this.EnableSideloading = $True }
			# -> install certificate (if needed)
			if ($BoundParameters['CertificatePath']) { $this.InstallCertificate = $True }
			# -> retrieve the package
			if ($BoundParameters['InstallFromWeb']) { $this.PackageSource = [PackageSource]'Download' }
			elseif ($BoundParameters['PackagePath']) { $this.PackageSource = [PackageSource]'Parameter' }
			else { $this.PackageSource = [PackageSource]::Folder }
			# -> install the app
			if ($BoundParameters['SingleUser']) { $this.InstallForSingleUser = $True }
			else { $this.InstallForAllUsers = $True }

			if ($BoundParameters['ForceInstall'] -and -not($this.NeedsInstall())) { $this.Reinstall = $True }
		}

		if ($BoundParameters['ForceLicense'] -or (-not $BoundParameters['SkipLicense'] -and -not (Test-AppWasActivated)))
		{
			# app not activated -> copy the license input file
			$this.CopyLicense = $True
		}

		if ($BoundParameters['UpdateStrategy'])
		{
			$this.CreateUpdateStrategy = $True
		}

		if (-not (Test-AdminPrivileges) -and -not (Test-SystemPrivileges))
		{
			if ($this.NeedsAdmin())
			{
				if (-not $BoundParameters['SkipElevate'])
				{
					$this.Elevate = $True
				}
				else
				{
					$this.FailElevate = $True # unable to get necessary privileges
				}
			}
		}
	}

	[bool] NeedsInstall() # check if the app is already installed in the required way (meaning the result of the operation specified by current state and parameters would make no difference)
	{
		$UserInstalledApp = Get-InstalledAppInfo
		#$AnyInstalledApp = Get-InstalledAppInfo -AllUsers $True
		$ProvisionedApp = $null
		if (Test-AdminPrivileges) { $ProvisionedApp = Get-ProvisionedAppInfo }
		if ($this.LaunchParameters['SingleUser'] -or $this.LaunchParameters['RegisterOnly']) # the app would be installed for current user
		{
			return -not [bool]$UserInstalledApp # is it not installed for the current user?
		}
		else # the app would be provisioned
		{
			return -not [bool]$ProvisionedApp # is it not provisioned?
		}
	}

	[bool] NeedsAdmin() # check if any planned operation requires administrative privileges
	{
		return ($this.EnableSideloading -or $this.EnableDataSharing -or $this.InstallForAllUsers)
	}

	[bool] HasNothingToDo()
	{
		# at least one of the essential steps has to be planned
		$SomethingToDo = ($this.EnableDataSharing -or					<# data sharing #>
			$this.RegisterForSingleUser -or								<# registration #>
			$this.InstallForSingleUser -or $this.InstallForAllUsers -or <# installation (including -ForceInstall) #>
			$this.CopyLicense -or										<# license/activation (including -ForceLicense) #>
			$this.CreateUpdateStrategy)									<# update bahaviour setting #>
		return -not $SomethingToDo
	}

	# pre-execution report
	[string] ToString()
	{
		$ExpectOpReport = 'Expected operation:'

		if ($this.HasNothingToDo())
		{
			if (Test-AdminPrivileges) { $AppInfo = (Get-InstalledAppInfo -AllUsers $True) }
			else { $AppInfo = (Get-InstalledAppInfo) }
			$ExpectOpReport += "`n  Nothing to do: "
			if ($AppInfo) { $ExpectOpReport += "The app is already installed (version: $($AppInfo.Version)), " }
			else { $ExpectOpReport += 'The installation would be skipped, ' }
			$ExpectOpReport += 'data sharing is enabled, '
			if (Test-AppWasActivated) { $ExpectOpReport += 'license does not need to be applied ' }
			else { $ExpectOpReport += 'license input would be skipped ' }
			$ExpectOpReport += 'and no new update strategy specified '
			$ExpectOpReport += '- the script will just exit.'
			return $ExpectOpReport
		}

		if ($this.FailElevate)
		{
			$ExpectOpReport += "`n  The script will fail because it won't be able to get administrative privileges."
			return $ExpectOpReport
		}

		if ($this.Elevate)
		{
			$ExpectOpReport += "`n  Script will try to relaunch itself with administrative privileges."
			if (-not $this.LaunchParameters['WhatIf'])
			{
				return $ExpectOpReport # everything else will be done by the new (relaunched) process
			}
			else
			{
				$ExpectOpReport += "`n`nThe relaunched script will do following operations (it might skip some of them due to additional information available to it):"
			}
		}

		if ($this.InstallCertificate) { $ExpectOpReport += "`n  The certificate will be installed." }
		if ($this.EnableSideloading) { $ExpectOpReport += "`n  Sideloading will be enabled." }
		if ($this.EnableDataSharing) { $ExpectOpReport += "`n  Data sharing will be enabled." }

		if ($this.PackageSource -eq [PackageSource]'Download') { $ExpectOpReport += "`n  The package and its dependencies will be downloaded to a temporary location." }

		if ($this.Reinstall) { $ExpectOpReport += "`n  The app will be reinstalled." }

		if ($this.InstallForAllUsers) { $ExpectOpReport += "`n  The app will be provisioned for all users on this machine." }
		if ($this.InstallForSingleUser) { $ExpectOpReport += "`n  The app will be installed for the current user." }
		if ($this.RegisterForSingleUser) { $ExpectOpReport += "`n  The app will be registered for the current user, if it is already installed on this machine."}

		if ($this.CopyLicense) { $ExpectOpReport += "`n  The app will be prepared to be activated during its next launch." }

		if ($this.CreateUpdateStrategy) { $ExpectOpReport += "`n  A new settings file specifying the update policy for the app will be created (possibly overwriting any existing update settings)." }

		if ($this.PackageSource -eq [PackageSource]'Download') { $ExpectOpReport += "`n  The downloaded temporary files will be deleted." }
		
		return $ExpectOpReport
	}

	# post-execution report
	[string] GetExecutionResultSummary()
	{
		$ExecutionSummary = 'Script execution result:'
		if ($this.InstallCertificate) { $ExecutionSummary += "`n  Certificate installation: " + (StepProgressToString($this.CertificateInstallProgress)) }
		if ($this.EnableDataSharing) { $ExecutionSummary += "`n  Data sharing setup: " + (StepProgressToString($this.DataSharingSetupProgress)) }
		if ($this.EnableSideloading) { $ExecutionSummary += "`n  Sideloading setup: " + (StepProgressToString($this.SideloadingSetupProgress)) }
		if ($this.InstallForSingleUser -or $this.InstallForAllUsers -or $this.RegisterForSingleUser) { $ExecutionSummary += "`n  App installation: " + (StepProgressToString($this.AppInstallationProgress)) }
		if ($this.CopyLicense) { $ExecutionSummary += "`n  License input setup: " + (StepProgressToString($this.LicenseInputSetupProgress)) }
		if ($this.CreateUpdateStrategy) { $ExecutionSummary += "`n  Update strategy setup: " + (StepProgressToString($this.UpdateStrategySetupProgress)) }
		return $ExecutionSummary
	}
}

function EncodeBase64($Text)
{
	$Bytes = [System.Text.Encoding]::UTF8.GetBytes($Text)
	return [Convert]::ToBase64String($Bytes)
}

 function DecodeBase64($Text)
 {
	return [System.Text.Encoding]::Unicode.GetString([System.Convert]::FromBase64String($Text))
 }

function StepProgressToString([StepProgress]$Progress)
{
	if ($Progress -eq [StepProgress]'NotExecuted') { return 'NOT EXECUTED' }
	if ($Progress -eq [StepProgress]'Success') { return 'SUCCEEDED' }
	if ($Progress -eq [StepProgress]'Failure') { return 'FAILED' }
}

# - script parameters check
function CheckParameters($BoundParameters, $UnboundArguments)
{
	$Valid = $True
	# check unrecognized params
	if ($UnboundArguments)
	{
		WriteEventLog Error 4000 "Unrecognized parameters: $UnboundArguments. Please run the script again with correct parameters. Exiting."
		$Valid = $False
	}
	# check valid enums
	if ($BoundParameters['UpdateStrategy'])
	{
		$OrigPolicy = $BoundParameters['UpdateStrategy']
		$BoundParameters['UpdateStrategy'] = [UpdateStrategy]::$OrigPolicy
		if (-not $BoundParameters['UpdateStrategy'])
		{
			WriteEventLog Error 4001 "Invalid value ($OrigPolicy) of the UpdateStrategy parameter. Use either User, Admin or Force."
			$Valid = $False
		}
	}
	# exit if not valid
	if (-not $Valid)
	{
		exit 4000
	}
}

# - script parameters (recognized) to list
function GetParameterList($BoundParameters)
{
	$ParamStrings = @()
	ForEach($Param in $BoundParameters.GetEnumerator())
	{
		if ($Param.Value -eq $True) # enabled switches
		{
			$ParamStrings += "-$($Param.Key)"
		}
		elseif ($Param.Value -ne $False) # parameters (paths)
		{
			$ParamStrings += "-$($Param.Key) "+'"'+$($Param.Value)+'"'
		}
	}
	return $ParamStrings
}
#endregion

#region DEBUGGING, HELP, REPORTS
# - help
function PrintHelp
{
	Write-Host "
Corinth Installation Script v$ScriptVersion

--------

String parameters:

- PackagePath
    Meaning: Path to the installation package. Any dependency packages are expected to be inside the `"Dependencies\x64`" folder right next to the package file.
    Default: The script will try to find a package within its folder (appx/msix/bundle right next to the script file) if no path specified.
    Failure action: If there is no package available on the specified path or if no path was specified and there is no single package in the default folder, the installation fails.
	
- LicensePath
    Meaning: Path to a license file which should be used to activate the app and unlock its full content during its first launch.
    Default: Will try to use a file named `"license_input`" located next to this script if no path specified.
    Failure action: If there is no file available on the specified path, the script will fail. This bahaviour can be overridden by using the SkipLicense option.
	
- UpdateStrategy
    Meaning: Species the preferred way of user interaction when handling updates. Allowed values: User, Admin or Force.
      User: Users are supposed to handle the updates himself: upon update discovery, a dialog is shown with the options to install the update or skip.
      Admin: A user is supposed to tell the admin that there is an update, and the admin should perform the update at an appropriate time.
      Force: The app updates itself automatically without any user interaction.
    Default: No settings file will be created, the app will use its default settings (User).
    Failure action: If an invalid option is specified (i.e. other than the above-mentioned three), the script fails, just as with any other unrecognized parameter.

--------

Switches:

- InstallFromWeb
    Meaning: The script will download latest version of the installation packages from the internet and install them. Overrides PackagePath parameter.
    Default: See PackagePath parameter.
	
- RegisterOnly
    Meaning: The script will only register (not install) the app for the current user. No further action will be taken. Presumes that the app is already installed on the computer.
    Default: If the switch is not provided, the app will be installed and related steps will be performed.
	
- SingleUser
    Meaning: Installs the app only for the executing user.
    Default: The app will be provisioned for all users on the computer if not otherwise specified.
	
- ForceInstall:
    Meaning: Will try to re-install the app even if it is already installed.
    Default: The installation and all the related steps will be skipped when there is any version of the app already available for current user.
	
- SkipLicense
    Meaning: Will not provide any license input. The LicensePath patameter will be ignored.
    Default: The script will try to copy a license input file to the app data, so that it can be activated. It will throw an error, if the license input is unavailable.
	
- ForceLicense
    Meaning: Will copy the license input file to the app data even if the app had been previously activated. Overrides SkipLicense.
    Default: If the script finds data sugesting that the app has been activated in the past, it will not try to provide the app with a license input file.
	
- Verbose
    Meaning: Enables console output. Useful when running the script manually.
    Default: Silent execution (except for the event logging and errors).
	
- Help: 
    Meaning: Prints this help and exits.
	
- PrintLog
    Meaning: Prints log and exits.
	
- PrintConfigReport
    Meaning: Prints configuration check report (i.e. Windows version, sideloading settings, installed version of the app etc.) and exits.
	
- WhatIf
    Meaning: Prints a summary of expected steps without actually executing them.
"
}

# - event log output
function PrintLog
{
	$Logs = Get-EventLog -LogName $EventLogName -Source $EventLogSource -ErrorAction SilentlyContinue
	if ($Logs.Count -gt 0)
	{
		Write-Output ($Logs | Format-Table | Out-String)
	}
	else
	{
		Write-Output 'No logs found.'
	}
}

# - configuration report
function GetTimestamp
{
	$Timestamp = Get-Date -UFormat '%Y/%m/%d %R'
	return $Timestamp
}

function GetWindowsSummary
{
	$WinVersionReg = Get-ItemProperty -Path "Registry::HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows NT\CurrentVersion"
	$WinProduct = $WinVersionReg.ProductName
	$WinEdition = $WinVersionReg.CompositionEditionID
	$WinVersion = $WinVersionReg.ReleaseId
	if ([System.Environment]::Is64BitOperatingSystem) # alternatively: (Get-CimInstance -ClassName win32_operatingsystem).OSArchitecture
	{
		$WinArchitecture = '64-bit'
	}
	else
	{
		$WinArchitecture = '32-bit'
	}
	return "Windows version: $WinProduct ($WinEdition) $WinArchitecture $WinVersion`n"
}

function CreateConfigReport($AppInstallationInfo)
{
	$ConfigReport = [System.Environment]::MachineName + ' ' + (GetTimestamp) + "`n" # header
	$ConfigReport += GetWindowsSummary
	<#
	$AppInstallerVersion = GetAppInstallerVersion
	if ($AppInstallerVersion)
	{
		$ConfigReport += "AppInstaller: True ($AppInstallerVersion)`n"
	}
	else
	{
		$ConfigReport += "AppInstaller: False`n"
	}
	#>
	$SideloadingEnabled = Test-Sideloading
	$ConfigReport += "Sideloading: $SideloadingEnabled`n"
	$SharingEnabled = Test-AppDataSharing
	$ConfigReport += "Data sharing: $SharingEnabled`n"
	$ConfigReport += "Installed version:`n"
	$ConfigReport += "  - this user ($Env:UserName): " + $AppInstallationInfo.AppInstalledVersionUser + "`n"
	$ConfigReport += '  - any user: ' + $AppInstallationInfo.AppInstalledVersionAny + "`n"
	$ConfigReport += '  - provisioned: ' + $AppInstallationInfo.AppProvisionedVersion + "`n"
	if ($AppInstallationInfo.AppWasInstalled)
	{
		if ($AppInstallationInfo.AppWasActivated) { $ConfigReport += "  - activated`n" }
		else { $ConfigReport += "  - not activated`n" }
	}
	return $ConfigReport
}

function SendMixpanelLaunchReport($ScriptInfo, $Parameters, $AppInstallationInfo)
{
	$DeviceID = [System.Environment]::MachineName
	$UserName = [System.Environment]::UserName
	$ParameterList = GetParameterList $Parameters
	$ReportObj = @{
		event = 'Script Started'
		properties = @{
			distinct_id = "$DeviceID ($UserName)"
			token = '4d4d27467a6f5dfe868e42b2218c1ab4'
			winProduct = Get-ItemPropertyValue -Path "Registry::HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows NT\CurrentVersion" -Name ProductName
			winVersion = Get-ItemPropertyValue -Path "Registry::HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows NT\CurrentVersion" -Name ReleaseId
			win64 = [System.Environment]::Is64BitOperatingSystem
			scriptVersion = $ScriptInfo.ScriptVersion
			scriptOrigin = $ScriptInfo.ScriptOrigin
			moduleVersion = $ScriptInfo.ModuleVersion
			moduleOrigin = $ScriptInfo.ModuleOrigin
			scriptParameters = $ParameterList
			isAdmin = Test-AdminPrivileges
			isSystem = Test-SystemPrivileges
			sideloadingEnabled = Test-Sideloading
			datasharingEnabled = Test-AppDataSharing
			appInstalledVersionUser = $AppInstallationInfo.AppInstalledVersionUser
			appInstalledVersionAny = $AppInstallationInfo.AppInstalledVersionUser
			appProvisionedVersion = $AppInstallationInfo.AppProvisionedVersion
		}
	}

	$ReportJson = $ReportObj |  ConvertTo-Json -Compress
	$ReportBase64 = EncodeBase64 $ReportJson
	$Response = Invoke-WebRequest -Uri "https://api.mixpanel.com/track/?data=$ReportBase64" -Method 'POST' -UseBasicParsing
	return $Response
}

function SendMixpanelError($ErrorID)
{
	$DeviceID = [System.Environment]::MachineName
	$UserName = [System.Environment]::UserName
	$ReportObj = @{
		event = "Error $ErrorID"
		properties = @{
			distinct_id = "$DeviceID ($UserName)"
			token = '4d4d27467a6f5dfe868e42b2218c1ab4'
		}
	}

	$ReportJson = $ReportObj |  ConvertTo-Json -Compress
	$ReportBase64 = EncodeBase64 $ReportJson
	$Response = Invoke-WebRequest -Uri "https://api.mixpanel.com/track/?data=$ReportBase64" -Method 'POST' -UseBasicParsing
	return $Response
}

function SendMixpanelExecutionSummary($ExecutionSummary)
{
	$DeviceID = [System.Environment]::MachineName
	$UserName = [System.Environment]::UserName
	$ReportObj = @{
		event = 'Script Finished'
		properties = @{
			distinct_id = "$DeviceID ($UserName)"
			token = '4d4d27467a6f5dfe868e42b2218c1ab4'
			executionSummary = $ExecutionSummary
		}
	}

	$ReportJson = $ReportObj |  ConvertTo-Json -Compress
	$ReportBase64 = EncodeBase64 $ReportJson
	$Response = Invoke-WebRequest -Uri "https://api.mixpanel.com/track/?data=$ReportBase64" -Method 'POST' -UseBasicParsing
	return $Response
}
#endregion

#region GENERAL PATH/FILE OPERATIONS
function ResolvePath($Path)
{
	$Path = Resolve-Path -Path $Path -ErrorAction SilentlyContinue # converts relative path to absolute
	if ($?) { return $Path.Path }
	else { return '' }
}

function CopyToPath($Source, $Destination)
{
	$DestinationDir = Split-Path -Parent $Destination
	if (-not (Test-Path $DestinationDir))
	{
		New-Item -Path $DestinationDir -ItemType Directory
	}
	Copy-Item $Source -Destination $Destination
	return $?
}
#endregion

#region PACKAGE PATH ACQUISITION
function GetPackagePathFromDir($PackageDir)
{
	$PackagePath = @()
	$PackagePath +=	Get-ChildItem (Join-Path $PackageDir "Corinth*.appx") | Where-Object { $_.Mode -NotMatch "d" }
	$PackagePath +=	Get-ChildItem (Join-Path $PackageDir "Corinth*.appxbundle") | Where-Object { $_.Mode -NotMatch "d" }
	$PackagePath +=	Get-ChildItem (Join-Path $PackageDir "Corinth*.msix") | Where-Object { $_.Mode -NotMatch "d" }
	$PackagePath +=	Get-ChildItem (Join-Path $PackageDir "Corinth*.msixbundle") | Where-Object { $_.Mode -NotMatch "d" }
	$PackageCount = [int]($PackagePath | Measure-Object).Count
	if ($PackageCount -lt 1)
	{
		WriteEventLog Error 4031 "No path to package provided and no packages found in the specified/default folder ($PackageDir). Exiting."
		exit 4031
	}
	$PackagePath = $PackagePath.FullName
	if ($PackageCount -gt 1)
	{
		WriteEventLog Error 4032 "No path to package provided and too many ($PackageCount) packages found in the specified/default folder ($PackageDir): $PackagePath. Exiting."
		exit 4032
	}

	return $PackagePath
}

function GetPackagePathDownload
{
	$PackagePath = Join-Path $env:tmp (Split-Path -Leaf $PackageUrl)
	WriteEventLog Information 1200 "Downloading app package from $PackageUrl to $PackagePath."
	Invoke-WebRequest -Uri $PackageUrl -OutFile $PackagePath
	if ($?)
	{
		WriteEventLog Information 1201 "Package $PackageUrl successfully downloaded."
	}
	else
	{
		WriteEventLog Error 4201 "Unable to download package $PackageUrl. Exiting."
		exit 4201
	}
	return $PackagePath
}
#endregion

#region DEPENDENCY PATHS ACQUISITION
function GetDependencyPathsDefault($PackageDir) # as taken from Add-AppDevPackage.ps1 (generated by VS during build)
{
	$DependencyPackagesDir = (Join-Path $PackageDir 'Dependencies')
	$DependencyPackagePaths = @()
	if (Test-Path $DependencyPackagesDir)
	{
		# Get architecture-neutral dependencies
		$DependencyPackagePaths += Get-ChildItem (Join-Path $DependencyPackagesDir "*.appx") | Where-Object { $_.Mode -NotMatch "d" }

		# Get architecture-specific dependencies
		if (($Env:Processor_Architecture -eq "x86" -or $Env:Processor_Architecture -eq "amd64") -and (Test-Path (Join-Path $DependencyPackagesDir "x86")))
		{
			$DependencyPackagePaths += Get-ChildItem (Join-Path $DependencyPackagesDir "x86\*.appx") | Where-Object { $_.Mode -NotMatch "d" }
		}
		if (($Env:Processor_Architecture -eq "amd64") -and (Test-Path (Join-Path $DependencyPackagesDir "x64")))
		{
			$DependencyPackagePaths += Get-ChildItem (Join-Path $DependencyPackagesDir "x64\*.appx") | Where-Object { $_.Mode -NotMatch "d" }
		}
		if (($Env:Processor_Architecture -eq "arm") -and (Test-Path (Join-Path $DependencyPackagesDir "arm")))
		{
			$DependencyPackagePaths += Get-ChildItem (Join-Path $DependencyPackagesDir "arm\*.appx") | Where-Object { $_.Mode -NotMatch "d" }
		}
	}
	if (Get-Member -InputObject $DependencyPackagePaths -name 'FullName' -MemberType Properties)
	{
		$DependencyPackagePaths = $DependencyPackagePaths.FullName
	}

	return $DependencyPackagePaths
}

function GetDependencyPathsDownload
{
	$DependencyDownloadTargets = ForEach ($Url in $DependencyUrls) { Join-Path $Env:Temp (Split-Path -Leaf $url) }
	$DependencyPackagePaths = @()
	for ($i = 0; $i -lt $DependencyUrls.Length; $i++)
	{
		$PackageUrl = $DependencyUrls[$i]
		Invoke-WebRequest -Uri $PackageUrl -OutFile $DependencyDownloadTargets[$i]
		if ($?) # download ok - use the dependency
		{
			$DependencyPackagePaths += $DependencyDownloadTargets[$i]
			WriteEventLog Information $(1211+$i) "Dependency package $PackageUrl successfully downloaded."
		}
		else # download failed - log, but try installing anyway
		{
			WriteEventLog Warning $(3301+$i) "Unable to download dependency package $PackageUrl. Will try to install the app anyway."
		}
	}
	return [string[]]$DependencyPackagePaths
}
#endregion

#region CLEANUP, EXIT
function Cleanup($CleanupPaths) # does not turn off sideloading again or uninstall the certificate
{
	if (-not $CleanupPaths)
	{
		WriteEventLog Information 2800 'Skipping cleanup - no files to remove.'
		return
	}
	WriteEventLog Information 1800 "Cleanup - removing the temporary files (downloaded package and dependencies): $CleanupPaths."

	ForEach ($Garbage in $CleanupPaths)
	{
		Remove-Item $Garbage #-Force
	}
	
	# check the cleanup result
	$CleanupLeftovers = @()
	ForEach ($Garbage in $CleanupPaths)
	{
		if (Test-Path $Garbage)
		{
			$CleanupLeftovers += $Garbage
		}
	}

	if ($CleanupLeftovers)
	{
		WriteEventLog Warning 3810 "Cleanup unsuccessful - following files left behind: $CleanupLeftovers. Consider removing them manually."
	}
	else
	{
		WriteEventLog Information 1810 "Cleanup successful."
	}
}

function LogProgressAndExit($ExitCode)
{
	# log result: modified config + applied setting
	$AppInstallationInfo = [AppInstallationInfo]::New()
	$ConfigReport = (CreateConfigReport $AppInstallationInfo) # summary of win version, enabled features etc.
	$ExecutionSummary = $ExecutionPlan.GetExecutionResultSummary()
	WriteEventLog Information 1900 "Script exiting.`n`n$ConfigReport`n$ExecutionSummary"
	SendMixpanelExecutionSummary $ExecutionSummary | Out-Null
	exit $ExitCode
}
#endregion

# --- MAIN OPERATION ---
function Main($MyInvocation)
{
	#region 0: PREPARATION PHASE
	# - 0.0 - log launch info (parameters, privileges)
	CreateEventLog # (try to) create event log source
	$ScriptInfo = [ScriptMetaInfo]::New($MyInvocation)
	$AppInstallationInfo = [AppInstallationInfo]::New()
	CheckParameters $MyInvocation.BoundParameters $MyInvocation.UnboundArguments # does not check the validity of their values (e.g. paths); exits if it finds unrecognized params
	$ParameterString = "$(GetParameterList $MyInvocation.BoundParameters)"
	$ScriptID = $ScriptInfo.GetScriptID()
	$LaunchReport = "$ScriptID running"
	if (Test-AdminPrivileges) { $LaunchReport += ' with admin privileges' }
	if (Test-SystemPrivileges) { $LaunchReport += ' in system context' }
	if ($ParameterString) { $LaunchReport += " with following parameters: $ParameterString" }
	$LaunchReport += '.'
	WriteEventLog Information 1000 $LaunchReport

	# TMP DEBUG - check the DISM availability
	if(Test-AdminPrivileges)
	{
		$dismFail = $False;
		Try
		{
			$tmp = Get-ProvisionedAppInfo;
			if ([bool]$tmp)
			{
				WriteEventLog Information 100 "DISM OK: app already provisioned"
			}
			else
			{
				WriteEventLog Information 200 "DISM OK: app not yet provisioned"
			}
		}
		Catch
		{
			$dismFail = $True;

			$ErrorMessage = $_.Exception.Message
			$FailedItem = $_.Exception.ItemName
			WriteEventLog Error 400 "Unable to get provisioning info.`n$ErrorMessage ($FailedItem)."
		}

		if ($dismFail)
		{
			# wait and try again
			$dismFail = $False;
			Try
			{
				Start-Sleep -Seconds 60
				$tmp = Get-ProvisionedAppInfo;
				if ([bool]$tmp)
				{
					WriteEventLog Information 101 "DISM OK after waiting: app already provisioned"
				}
				else
				{
					WriteEventLog Information 201 "DISM OK after waiting: app not yet provisioned"
				}
			}
			Catch
			{
				$dismFail = $True;
	
				$ErrorMessage = $_.Exception.Message
				$FailedItem = $_.Exception.ItemName
				WriteEventLog Error 401 "Unable to get provisioning info even after waiting.`n$ErrorMessage ($FailedItem)."
			}	
		}

		if ($dismFail)
		{
			# import explicitly and try again
			Try
			{
				Import-Module DISM
				Start-Sleep -Seconds 30

				$tmp = Get-ProvisionedAppInfo;
				if ([bool]$tmp)
				{
					WriteEventLog Information 102 "DISM OK after import: app already provisioned"
				}
				else
				{
					WriteEventLog Information 202 "DISM OK after import: app not yet provisioned"
				}
			}
			Catch
			{
				$dismFail = $True;
	
				$ErrorMessage = $_.Exception.Message
				$FailedItem = $_.Exception.ItemName
				WriteEventLog Error 402 "Unable to get provisioning info even after import.`n$ErrorMessage ($FailedItem)."
			}	
		}
	}

	# - 0.1 - log current configuration and expected execution info
	$ExecutionPlan = [ScriptExecutionPlan]::New($MyInvocation.BoundParameters)
	$ExpectedOpReport = $ExecutionPlan.ToString() # summary of the expected behaiour of this script
	if ($WhatIf)
	{
		Write-Output $ExpectedOpReport
		exit 0
	}
	
	$ConfigReport = (CreateConfigReport $AppInstallationInfo) # summary of win version, enabled features etc.

	WriteEventLog Information 1010 "$ConfigReport`n$ExpectedOpReport" # log configuration report and expected operations

	if ($ExecutionPlan.HasNothingToDo())
	{
		WriteEventLog Information 2010 'Nothing to do given the current cofiguration and script parameters. Exiting.'
		exit 0
	}

	# - 0.2 - check OS
	# $WinVersion = [int](Get-ItemPropertyValue -Path "Registry::HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows NT\CurrentVersion" -Name ReleaseId) # convert to int for correct comparison
	# if ($WinVersion -lt 1803 -or -not ([System.Environment]::Is64BitOperatingSystem))
	if (-not ([System.Environment]::Is64BitOperatingSystem)) # fix: version does not have to be an integer => let's skip the check as nobody uses such an old version anymore anyway
	{
		WriteEventLog Warning 3020 "The installation will probably not be successful on this OS. You should use Windows 10 64-bit, version 1803 or newer to install and run Corinth 3D."
	}
	else
	{
		WriteEventLog Information 1020 "Operating system check OK."
	}

	# - 0.3 - pre-process and check parameters (paths)
	if ($ExecutionPlan.InstallForAllUsers -or $ExecutionPlan.InstallForSingleUser)
	{
		if ($ExecutionPlan.PackageSource -eq [PackageSource]'Parameter') # specified by parameter
		{
			$OrigPackagePath = $PackagePath
			if (Test-Path $PackagePath -PathType Container) # path to a directory (not a file) was specified
			{
				$PackagePath = GetPackagePathFromDir $PackagePath # try to get a package from that dir - exits if it fails to find a package
			}
			$PackagePath = ResolvePath $PackagePath # converts relative path to absolute
			$MyInvocation.BoundParameters.PackagePath = $PackagePath # update parsed params for relaunch
			if (-not $PackagePath -or -not (Test-Path $PackagePath)) # was not a valid path
			{
				WriteEventLog Error 4030 "The specified package path ($OrigPackagePath) is not valid. Exiting."
				LogProgressAndExit 4030
			}
		}
		elseif ($ExecutionPlan.PackageSource -eq [PackageSource]'Folder') # try looking next to the script (default location)
		{
			$PackagePath = GetPackagePathFromDir $ScriptInfo.ScriptDir # exits if it fails to find a package (exactly one package)
			$MyInvocation.BoundParameters.PackagePath = $PackagePath # update parsed params for relaunch
		}
	}
	
	if ($ExecutionPlan.CopyLicense) # specified by parameter or next to the script (default param value)
	{
		if (-not $LicensePath) { $LicensePath = Join-Path $ScriptInfo.ScriptDir 'license_input' } # default location is next to the script
		$OrigLicensePath = $LicensePath
		$LicensePath = ResolvePath $LicensePath # converts relative path to absolute
		$MyInvocation.BoundParameters.LicensePath = $LicensePath # update parsed params for relaunch
		if (-not $LicensePath -or -not (Test-Path $LicensePath)) # was not a valid path
		{
			WriteEventLog Error 4033 "The specified license path ($OrigLicensePath) is not valid. Exiting."
			LogProgressAndExit 4033
		}
	}
	
	if ($ExecutionPlan.InstallCertificate)
	{
		$OrigCertificatePath = $CertificatePath
		$CertificatePath = ResolvePath $CertificatePath # converts relative path to absolute
		$MyInvocation.BoundParameters.CertificatePath = $CertificatePath # update parsed params for relaunch
		if (-not $CertificatePath -or -not (Test-Path $CertificatePath)) # was not a valid path
		{
			WriteEventLog Warning 3034 "The specified certificate path ($OrigCertificatePath) is not valid and will be skipped."
			$CertificatePath = $null
			$MyInvocation.BoundParameters.Remove('CertificatePath') | Out-Null
			$ExecutionPlan.InstallCertificate = $False
		}
	}
	#endregion
	
	#region 1: OBTAIN NECESSARY PRIVILEGES (RELAUNCH)
	if ($ExecutionPlan.Elevate)
	{
		WriteEventLog Information 1100 'Relaunching with administrative privileges.'
		
		$ScriptPath = (Get-Variable MyInvocation).Value.MyCommand.Path
		[string[]]$RelaunchArgs = @('-NoProfile', '-ExecutionPolicy Unrestricted', '-File "'+$ScriptPath+'"', '-SkipElevate')
		$RelaunchArgs += "$(GetParameterList $MyInvocation.BoundParameters)" # parameters after the path modification
		# get the right path to the module
		if ($MyInvocation.BoundParameters['ModulePath'])
		{
			$ModulePath = Resolve-Path $MyInvocation.BoundParameters['ModulePath']
		}
		else
		{
			$ModulePath = Resolve-Path '.\CCInstallUtils.psd1'
		}
		$RelaunchArgs += ('-ModulePath "'+$ModulePath+'"')
		$NewProcess = Start-Process "$PsHome\PowerShell.exe" -Verb RunAs -ArgumentList $RelaunchArgs -PassThru -Wait # waits until the script finishes (=> can be used synchronously)
		exit $NewProcess.ExitCode # TODO: forward output
	}

	if (-not $ExecutionPlan.NeedsAdmin() -or (Test-AdminPrivileges) -or (Test-SystemPrivileges))
	{
		WriteEventLog Information 2100 'Script running with the necessary privileges.'
		SendMixpanelLaunchReport $ScriptInfo $MyInvocation.BoundParameters $AppInstallationInfo | Out-Null
	}
	else
	{
		WriteEventLog Error 4100 'Unable to obtain necessary privileges. Exiting.'
		LogProgressAndExit 4100
	}

	# set own priority to "High"
	$Process = Get-Process -Id $pid
	$Process.PriorityClass = "High"
	#endregion
	
	#region 2: GET INSTALLATION PACKAGES
	$CleanupPaths = @()
	if ($ExecutionPlan.InstallForAllUsers -or $ExecutionPlan.InstallForSingleUser)
	{
		# - 2.0 - app package
		if ($ExecutionPlan.PackageSource -ne [PackageSource]'Download')
		{
			$InstallPackagePath = $PackagePath # already set during 0.2
		}
		else # package from the web
		{
			$InstallPackagePath = GetPackagePathDownload # downloads the package
			$CleanupPaths += $InstallPackagePath
		}

		# - 2.1 - dependencies
		if ($ExecutionPlan.PackageSource -ne [PackageSource]'Download')
		{
			$PackageDir = Split-Path -Parent $PackagePath
			$InstallDependencyPaths = GetDependencyPathsDefault $PackageDir
		}
		else
		{
			$InstallDependencyPaths = GetDependencyPathsDownload
			$CleanupPaths += $InstallDependencyPaths
		}

		# - 2.2 - log the package source
		if ($ExecutionPlan.PackageSource -eq [PackageSource]'Parameter')
		{
			WriteEventLog Information 1220 "Package provided by the parameter will be installed ($InstallPackagePath)."
		}
		elseif ($ExecutionPlan.PackageSource -eq [PackageSource]'Download') 
		{
			WriteEventLog Information 1221 "Downloaded package will be installed ($InstallPackagePath)."
		}
		else # [PackageSource]'Folder' - default
		{
			WriteEventLog Information 1222 "No package source provided via parameter, a package from the default location will be used ($InstallPackagePath)."
		}

		# - 2.3 - log dependecies
		if ($InstallDependencyPaths -and $InstallDependencyPaths.Count -gt 0)
		{
			WriteEventLog Information 1230 "Following dependencies will be installed: $InstallDependencyPaths."
		}
		else
		{
			WriteEventLog Warning 3230 'No dependecies found. Have you placed them in the correct folder?'
		}
	}
	#endregion
	
	#region 3: CHECK CERTIFICATE
	if ($ExecutionPlan.InstallCertificate) # only for debug purposes - the production package should be signed by a trusted certificate
	{
		if (Install-Certificate $CertificatePath)
		{
			WriteEventLog Information 1300 "The certificate should be now installed ($CertificatePath)."
			$ExecutionPlan.CertificateInstallProgress = [StepProgress]'Success'
		}
		else 
		{
			WriteEventLog Warning 3300 "Unable to install the certificate ($CertificatePath). Trying to install the app anyway."
			$ExecutionPlan.CertificateInstallProgress = [StepProgress]'Failure'
		}
	}
	
	if ($ExecutionPlan.InstallForAllUsers -or $ExecutionPlan.InstallForSingleUser)
	{
		if (Test-PackageSignature $InstallPackagePath) # if not signed by a valid certificate, the installation fails
		{
			WriteEventLog Information 1310 "Package certificate OK."
		}
		else
		{
			WriteEventLog Error 4310 "Package certificate not valid. Exiting."
			Cleanup $CleanupPaths
			LogProgressAndExit 4310
		}
	}
	#endregion
	
	#region 4: ENABLE SIDELOADING
	if ($ExecutionPlan.EnableSideloading)
	{
		WriteEventLog Information 1400 'Turning sideloading on.'
		Enable-Sideloading | Out-Null
		if ($? -and (Test-Sideloading))
		{
			WriteEventLog Information 1410 'Sideloading enabled.'
			$ExecutionPlan.SideloadingSetupProgress = [StepProgress]'Success'
		}
		else
		{
			WriteEventLog Error 4410 'Unable to turn sideloading on. Exiting.'
			$ExecutionPlan.SideloadingSetupProgress = [StepProgress]'Failure'
			exit 4410
		}
	}
	else
	{
		WriteEventLog Information 2400 'Skipping sideloading setup.'
	}
	#endregion

	#region 5: ENABLE DATA SHARING
	if ($ExecutionPlan.EnableDataSharing)
	{
		WriteEventLog Information 1500 'Turning data sharing on.'
		Enable-AppDataSharing | Out-Null
		if ($? -and (Test-AppDataSharing))
		{
			WriteEventLog Information 1510 'Data sharing enabled.'
			$ExecutionPlan.DataSharingSetupProgress = [StepProgress]'Success'
		}
		else
		{
			WriteEventLog Error 4510 'Unable to turn data sharing on. Exiting.'
			$ExecutionPlan.DataSharingSetupProgress = [StepProgress]'Failure'
			exit 4510
		}
	}
	else
	{
		WriteEventLog Information 2500 'Skipping data sharing setup.'
	}
	#endregion
	
	#region 6: REGISTER/INSTALL/PROVISION APP
	if ($ExecutionPlan.InstallForAllUsers -or $ExecutionPlan.InstallForSingleUser -or $ExecutionPlan.RegisterForSingleUser)
	{
		if ($ExecutionPlan.InstallForAllUsers) # for current and "new" users (this seems to add the app for all the existing users as well as for those added in the future)
		{
			$PreviousVersion = if (Get-ProvisionedAppInfo) { (Get-ProvisionedAppInfo).Version } else { $null }
			WriteEventLog Information 1600 'Starting provisioning.'
			if ($Verbose)
			{
				Write-Host 'Provisioning. This might take a while, please wait.' # no loading bar, sadly
			}
			$InstallSuccess = Install-AppxPackage $InstallPackagePath $InstallDependencyPaths -AllUsers $True
		}
		else # for the current user only
		{
			$PreviousVersion = if (Get-InstalledAppInfo) { (Get-InstalledAppInfo).Version } else { $null }
			if ($ExecutionPlan.InstallForSingleUser) {
				WriteEventLog Information 1610 'Starting installation.' 
				$InstallSuccess = Install-AppxPackage $InstallPackagePath $InstallDependencyPaths
			}
			else { # register only
				WriteEventLog Information 1620 'Starting registration.'
				$InstallSuccess = Register-AppxPackage
			}
		}

		# check the result
		if ($ExecutionPlan.InstallForAllUsers) { $CurrentVersion = if (Get-ProvisionedAppInfo) { (Get-ProvisionedAppInfo).Version } else { $null } }
		else { $CurrentVersion = if (Get-InstalledAppInfo) { (Get-InstalledAppInfo).Version } else { $null } }

		if ($InstallSuccess)
		{
			$ExecutionPlan.AppInstallationProgress = [StepProgress]'Success'
			$Timer = 0
			while ($Timer -lt 60 -and $CurrentVersion -eq $PreviousVersion) # this always waits for full 60 seconds when forcibly re-installing the same version
			{
				Start-Sleep -Seconds 3 # wait a bit for the Windows to process it
				$Timer += 3
				if ($ExecutionPlan.InstallForAllUsers) { $CurrentVersion = if (Get-ProvisionedAppInfo) { (Get-ProvisionedAppInfo).Version } else { $null } }
				else { $CurrentVersion = if (Get-InstalledAppInfo) { (Get-InstalledAppInfo).Version } else { $null } }
			}

			<#
			if ($ExecutionPlan.InstallForAllUsers -and $CurrentVersion -and -not (Get-InstalledAppInfo)) # successfully provisioned, but not available for the current user
			{
				Register-AppxPackage
			}
			#>

			if ($CurrentVersion)
			{
				if (-not $PreviousVersion) { WriteEventLog Information 1640 "App installed, version: $CurrentVersion." }
				else { WriteEventLog Information 1641 "App upgraded, from version $PreviousVersion to version $CurrentVersion." }
			}
			else
			{
				WriteEventLog Error 4640 'Installation succeeded, but the app is still not available. Exiting.' # should not happen
				Cleanup $CleanupPaths
				LogProgressAndExit 4640
			}
		}
		else
		{
			$ExecutionPlan.AppInstallationProgress = [StepProgress]'Failure'
			WriteEventLog Error 4641 'Installation failed. Exiting.'
			Cleanup $CleanupPaths
			LogProgressAndExit 4641
		}
	}
	else
	{
		WriteEventLog Information 2640 'Skipping installation.'
		# check existing - app still needed to continue (i.e. apply the license input and udpate settings)
		$CurrentVersion = if (Get-InstalledAppInfo -AllUsers $True) { (Get-InstalledAppInfo -AllUsers $True).Version } else { $null } 
		if (-not $CurrentVersion)
		{
			WriteEventLog Error 4642 'App installation not found. Exiting.'
			Cleanup $CleanupPaths
			LogProgressAndExit 4642
		}
	}
	#endregion

	#region 7: PROVIDE LICENSE AND UPDATE SETTINGS
	if ($ExecutionPlan.CopyLicense -or $ExecutionPlan.CreateUpdateStrategy) 
	{
		$InstalledPackageInfo = if (Test-AdminPrivileges) { Get-InstalledAppInfo -AllUsers $True } else { Get-InstalledAppInfo }
		$SharedPath = Get-SharedDataPath # is shared and accessible by the app, but not created until the first launch
		$InstallPath = Get-InstallPath -PackageInfo $InstalledPackageInfo # is shared, accessible by the app and available right after installation, but is supposed to be "read-only"
		$LocalPath = Get-LocalDataPath # is accessible and available, but not shared
		# might be a while until the local data folder appears after installation
		$timer = 0
		while ($timer -lt 60 -and (Get-InstalledAppInfo) -and -not $LocalPath)
		{
			Start-Sleep -Seconds 3 # wait a bit for the Windows to process it
			$Timer += 3
			$LocalPath = Get-LocalDataPath
		}
	}

	# - 7.0 - license input
	if ($ExecutionPlan.CopyLicense)
	{
		$LicSuccess = $False
		if (Test-SystemPrivileges)
		{
			if ($InstallPath)
			{
				$LicenseTargetPath = Join-Path $InstallPath 'license_input'
				$LicSuccess = (CopyToPath $LicensePath $LicenseTargetPath -and Test-Path $LicenseTargetPath)
			}
			if ($LicSuccess)
			{
				WriteEventLog Information 1700 'License input copied to a shared folder - the app can be activated by any user.'
				$ExecutionPlan.LicenseInputSetupProgress = [StepProgress]'Success'
			}
			else
			{
				WriteEventLog Warning 3700 'Unable to copy the license input to the shared folder - trying local user folder.'
				#$ExecutionPlan.LicenseInputSetupProgress = [StepProgress]'Failure'
			}
		}

		if (-not $LicSuccess)
		{
			if ($LocalPath)
			{
				$LicenseTargetPath = Join-Path $LocalPath 'license_input'
				$LicSuccess = (CopyToPath $LicensePath $LicenseTargetPath -and Test-Path $LicenseTargetPath)
			}
			if ($LicSuccess)
			{
				WriteEventLog Information 1701 "License input copied to a local user folder - the app can be activated by the current user ($Env:UserName)."
				$ExecutionPlan.LicenseInputSetupProgress = [StepProgress]'Success'
			}
			else
			{
				WriteEventLog Error 4701 'Unable to copy the license input to local user folder - the app will not be activated automatically during launch, consider manual activation.'
				$ExecutionPlan.LicenseInputSetupProgress = [StepProgress]'Failure'
			}
		}
	}
	else
	{
		WriteEventLog Information 2700 'Skipping license input.'
	}

	# - 7.1 - update strategy
	if ($ExecutionPlan.CreateUpdateStrategy)
	{
		# determine xml content
		$UpdateStrategyXmlBegin = "<?xml version=`"1.0`"?>`n<UpdateSettings>`n  <Strategy>"
		$UpdateStrategyXmlEnd = "</Strategy> <!-- User, Admin, Force -->`n</UpdateSettings>"

		if ($UpdateStrategy -eq [UpdateStrategy]::'User')
		{
			$UpdateStrategyXmlContent = $UpdateStrategyXmlBegin+'User'+$UpdateStrategyXmlEnd
		}
		elseif ($UpdateStrategy -eq [UpdateStrategy]::'Force')
		{
			$UpdateStrategyXmlContent = $UpdateStrategyXmlBegin+'Force'+$UpdateStrategyXmlEnd
		}
		else # [UpdateStrategy]::'Admin' - default
		{
			$UpdateStrategyXmlContent = $UpdateStrategyXmlBegin+'Admin'+$UpdateStrategyXmlEnd
		}
		# get destination path
		$AppDataPath = $null
		if ($SharedPath)
		{
			$AppDataPath = $SharedPath
		}
		elseif ($InstallPath -and (Test-SystemPrivileges))
		{
			$AppDataPath = $InstallPath
		}
		# create the file
		$UPSuccess = $False
		if ($AppDataPath)
		{
			$UpdateStrategyPath = Join-Path $AppDataPath 'Settings.xml'
			if (-not (Test-Path $UpdateStrategyPath))
			{
				New-Item $UpdateStrategyPath | Out-Null
			}
			Set-Content $UpdateStrategyPath $UpdateStrategyXmlContent
			$UPSuccess = $?
		}
		# check the result
		if ($UPSuccess)
		{
			WriteEventLog Information 1710 "Update settings successfully created ($UpdateStrategyPath)."
			$ExecutionPlan.UpdateStrategySetupProgress = [StepProgress]'Success'
		}
		else
		{
			WriteEventLog Error 4710 'Unable to create update settings file. Please consider creating it manually in the shared app data folder.'
			$ExecutionPlan.UpdateStrategySetupProgress = [StepProgress]'Failure'
		}
	}
	else
	{
		WriteEventLog Information 2710 'Skipping update settings creation.'	
	}
	#endregion
	
	#region 8: CLEAN UP
	Cleanup $CleanupPaths
	#endregion
	
	#region 9: EVALUATE EXECUTION
	LogProgressAndExit 0
	#endregion
}

# - print help
if ($Help)
{
	PrintHelp
	exit 0
}

# - print event logs
if ($PrintLog)
{
	PrintLog
	exit 0
}

if ($PrintConfigReport) # summary of win version, enabled features etc.
{
	$AppInstallationInfo = [AppInstallationInfo]::New()
	Write-Output $(CreateConfigReport $AppInstallationInfo)
	exit 0
}

Main $MyInvocation

# SIG # Begin signature block
# MIIfHQYJKoZIhvcNAQcCoIIfDjCCHwoCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUxAVrzvukEFHInu2Fju4Bfdyv
# d12gghkEMIIFDDCCA/SgAwIBAgIRAJ0ixooo0boZelXoKg5DAFQwDQYJKoZIhvcN
# AQELBQAwfDELMAkGA1UEBhMCR0IxGzAZBgNVBAgTEkdyZWF0ZXIgTWFuY2hlc3Rl
# cjEQMA4GA1UEBxMHU2FsZm9yZDEYMBYGA1UEChMPU2VjdGlnbyBMaW1pdGVkMSQw
# IgYDVQQDExtTZWN0aWdvIFJTQSBDb2RlIFNpZ25pbmcgQ0EwHhcNMjEwNDA2MDAw
# MDAwWhcNMjIwMzI0MjM1OTU5WjBPMQswCQYDVQQGEwJDWjEOMAwGA1UEBwwFUHJh
# aGExFzAVBgNVBAoMDkNvcmludGggcy5yLm8uMRcwFQYDVQQDDA5Db3JpbnRoIHMu
# ci5vLjCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBANCSI5UhlyFsK0/1
# OsfO3rajcKGxMp/Qo/uUs74BI5IaZ+kSRDgvzI8Ay9/O/x3EiO6gW1jDMbpOhHHR
# x72DmDavSP4JTvkTaPvwhc7eoCGCsyhi8J1Kca4wtrpUushcfaBn7CS9RVX6rDIB
# uTPsyVuPaMCjj/f1AWB+tE+4Xd1LVSPa+rs7akopoMmtrZ9jA7bDavclgAUY83jg
# BvRhqSmawCglc4o/Ap21iPvaUDfDGwPTz5ff6O/sKc+dFmpHH6RSD/zIy5vJ/2lm
# 5t2o11Oe8f3HshraKbemmLXHmFeoBJuySzWmlrCD/Q4jDRlkAeCa0y6Ttcu1Q81g
# Md/60o0CAwEAAaOCAbQwggGwMB8GA1UdIwQYMBaAFA7hOqhTOjHVir7Bu61nGgOF
# rTQOMB0GA1UdDgQWBBR9VxGgGc7reAaWPimwyFWfXcH3jzAOBgNVHQ8BAf8EBAMC
# B4AwDAYDVR0TAQH/BAIwADATBgNVHSUEDDAKBggrBgEFBQcDAzARBglghkgBhvhC
# AQEEBAMCBBAwSgYDVR0gBEMwQTA1BgwrBgEEAbIxAQIBAwIwJTAjBggrBgEFBQcC
# ARYXaHR0cHM6Ly9zZWN0aWdvLmNvbS9DUFMwCAYGZ4EMAQQBMEMGA1UdHwQ8MDow
# OKA2oDSGMmh0dHA6Ly9jcmwuc2VjdGlnby5jb20vU2VjdGlnb1JTQUNvZGVTaWdu
# aW5nQ0EuY3JsMHMGCCsGAQUFBwEBBGcwZTA+BggrBgEFBQcwAoYyaHR0cDovL2Ny
# dC5zZWN0aWdvLmNvbS9TZWN0aWdvUlNBQ29kZVNpZ25pbmdDQS5jcnQwIwYIKwYB
# BQUHMAGGF2h0dHA6Ly9vY3NwLnNlY3RpZ28uY29tMCIGA1UdEQQbMBmBF2RldmVs
# b3BlckBjb3JpbnRoM2QuY29tMA0GCSqGSIb3DQEBCwUAA4IBAQBPF0WmYhSqERFn
# q4B2EwlPbItCuQxOUBdVD8Z485t66uDfxs2wpJaHIfUdx/6sQahqt9vtwqNdmaIH
# JCV4cwfgQrEbxo7kuKt+Q11ZCsEqbCNsmfT4T2s3JBFCF8QpW/VfZP+DGT2mF+TB
# TG/0+/ZERTDXWVm26CBWTMTnUXkROV9Hwdym48hPx5WnkngqSa+byYJAS++dbUxf
# fl3+kYE2JHHLxguqbFQVbEluM3urkv4uAraoQCTB15Di1NuHeMkgVpDn6Q3nRdEf
# 0sHDLc/I1LPOuxwgGAW3gPkik/W8nsbl0a3n2d4EYRKaB9tJNlmI5c5XiE8BVnAb
# 8UdevLa3MIIF9TCCA92gAwIBAgIQHaJIMG+bJhjQguCWfTPTajANBgkqhkiG9w0B
# AQwFADCBiDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCk5ldyBKZXJzZXkxFDASBgNV
# BAcTC0plcnNleSBDaXR5MR4wHAYDVQQKExVUaGUgVVNFUlRSVVNUIE5ldHdvcmsx
# LjAsBgNVBAMTJVVTRVJUcnVzdCBSU0EgQ2VydGlmaWNhdGlvbiBBdXRob3JpdHkw
# HhcNMTgxMTAyMDAwMDAwWhcNMzAxMjMxMjM1OTU5WjB8MQswCQYDVQQGEwJHQjEb
# MBkGA1UECBMSR3JlYXRlciBNYW5jaGVzdGVyMRAwDgYDVQQHEwdTYWxmb3JkMRgw
# FgYDVQQKEw9TZWN0aWdvIExpbWl0ZWQxJDAiBgNVBAMTG1NlY3RpZ28gUlNBIENv
# ZGUgU2lnbmluZyBDQTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAIYi
# jTKFehifSfCWL2MIHi3cfJ8Uz+MmtiVmKUCGVEZ0MWLFEO2yhyemmcuVMMBW9aR1
# xqkOUGKlUZEQauBLYq798PgYrKf/7i4zIPoMGYmobHutAMNhodxpZW0fbieW15dR
# hqb0J+V8aouVHltg1X7XFpKcAC9o95ftanK+ODtj3o+/bkxBXRIgCFnoOc2P0tbP
# BrRXBbZOoT5Xax+YvMRi1hsLjcdmG0qfnYHEckC14l/vC0X/o84Xpi1VsLewvFRq
# nbyNVlPG8Lp5UEks9wO5/i9lNfIi6iwHr0bZ+UYc3Ix8cSjz/qfGFN1VkW6KEQ3f
# BiSVfQ+noXw62oY1YdMCAwEAAaOCAWQwggFgMB8GA1UdIwQYMBaAFFN5v1qqK0rP
# VIDh2JvAnfKyA2bLMB0GA1UdDgQWBBQO4TqoUzox1Yq+wbutZxoDha00DjAOBgNV
# HQ8BAf8EBAMCAYYwEgYDVR0TAQH/BAgwBgEB/wIBADAdBgNVHSUEFjAUBggrBgEF
# BQcDAwYIKwYBBQUHAwgwEQYDVR0gBAowCDAGBgRVHSAAMFAGA1UdHwRJMEcwRaBD
# oEGGP2h0dHA6Ly9jcmwudXNlcnRydXN0LmNvbS9VU0VSVHJ1c3RSU0FDZXJ0aWZp
# Y2F0aW9uQXV0aG9yaXR5LmNybDB2BggrBgEFBQcBAQRqMGgwPwYIKwYBBQUHMAKG
# M2h0dHA6Ly9jcnQudXNlcnRydXN0LmNvbS9VU0VSVHJ1c3RSU0FBZGRUcnVzdENB
# LmNydDAlBggrBgEFBQcwAYYZaHR0cDovL29jc3AudXNlcnRydXN0LmNvbTANBgkq
# hkiG9w0BAQwFAAOCAgEATWNQ7Uc0SmGk295qKoyb8QAAHh1iezrXMsL2s+Bjs/th
# AIiaG20QBwRPvrjqiXgi6w9G7PNGXkBGiRL0C3danCpBOvzW9Ovn9xWVM8Ohgyi3
# 3i/klPeFM4MtSkBIv5rCT0qxjyT0s4E307dksKYjalloUkJf/wTr4XRleQj1qZPe
# a3FAmZa6ePG5yOLDCBaxq2NayBWAbXReSnV+pbjDbLXP30p5h1zHQE1jNfYw08+1
# Cg4LBH+gS667o6XQhACTPlNdNKUANWlsvp8gJRANGftQkGG+OY96jk32nw4e/gdR
# EmaDJhlIlc5KycF/8zoFm/lv34h/wCOe0h5DekUxwZxNqfBZslkZ6GqNKQQCd3xL
# S81wvjqyVVp4Pry7bwMQJXcVNIr5NsxDkuS6T/FikyglVyn7URnHoSVAaoRXxrKd
# sbwcCtp8Z359LukoTBh+xHsxQXGaSynsCz1XUNLK3f2eBVHlRHjdAd6xdZgNVCT9
# 8E7j4viDvXK6yz067vBeF5Jobchh+abxKgoLpbn0nu6YMgWFnuv5gynTxix9vTp3
# Los3QqBqgu07SqqUEKThDfgXxbZaeTMYkuO1dfih6Y4KJR7kHvGfWocj/5+kUZ77
# OYARzdu1xKeogG/lU9Tg46LC0lsa+jImLWpXcBw8pFguo/NbSwfcMlnzh6cabVgw
# ggbsMIIE1KADAgECAhAwD2+s3WaYdHypRjaneC25MA0GCSqGSIb3DQEBDAUAMIGI
# MQswCQYDVQQGEwJVUzETMBEGA1UECBMKTmV3IEplcnNleTEUMBIGA1UEBxMLSmVy
# c2V5IENpdHkxHjAcBgNVBAoTFVRoZSBVU0VSVFJVU1QgTmV0d29yazEuMCwGA1UE
# AxMlVVNFUlRydXN0IFJTQSBDZXJ0aWZpY2F0aW9uIEF1dGhvcml0eTAeFw0xOTA1
# MDIwMDAwMDBaFw0zODAxMTgyMzU5NTlaMH0xCzAJBgNVBAYTAkdCMRswGQYDVQQI
# ExJHcmVhdGVyIE1hbmNoZXN0ZXIxEDAOBgNVBAcTB1NhbGZvcmQxGDAWBgNVBAoT
# D1NlY3RpZ28gTGltaXRlZDElMCMGA1UEAxMcU2VjdGlnbyBSU0EgVGltZSBTdGFt
# cGluZyBDQTCCAiIwDQYJKoZIhvcNAQEBBQADggIPADCCAgoCggIBAMgbAa/ZLH6I
# mX0BmD8gkL2cgCFUk7nPoD5T77NawHbWGgSlzkeDtevEzEk0y/NFZbn5p2QWJgn7
# 1TJSeS7JY8ITm7aGPwEFkmZvIavVcRB5h/RGKs3EWsnb111JTXJWD9zJ41OYOioe
# /M5YSdO/8zm7uaQjQqzQFcN/nqJc1zjxFrJw06PE37PFcqwuCnf8DZRSt/wflXMk
# PQEovA8NT7ORAY5unSd1VdEXOzQhe5cBlK9/gM/REQpXhMl/VuC9RpyCvpSdv7Qg
# sGB+uE31DT/b0OqFjIpWcdEtlEzIjDzTFKKcvSb/01Mgx2Bpm1gKVPQF5/0xrPnI
# hRfHuCkZpCkvRuPd25Ffnz82Pg4wZytGtzWvlr7aTGDMqLufDRTUGMQwmHSCIc9i
# VrUhcxIe/arKCFiHd6QV6xlV/9A5VC0m7kUaOm/N14Tw1/AoxU9kgwLU++Le8bwC
# KPRt2ieKBtKWh97oaw7wW33pdmmTIBxKlyx3GSuTlZicl57rjsF4VsZEJd8GEpoG
# LZ8DXv2DolNnyrH6jaFkyYiSWcuoRsDJ8qb/fVfbEnb6ikEk1Bv8cqUUotStQxyk
# SYtBORQDHin6G6UirqXDTYLQjdprt9v3GEBXc/Bxo/tKfUU2wfeNgvq5yQ1TgH36
# tjlYMu9vGFCJ10+dM70atZ2h3pVBeqeDAgMBAAGjggFaMIIBVjAfBgNVHSMEGDAW
# gBRTeb9aqitKz1SA4dibwJ3ysgNmyzAdBgNVHQ4EFgQUGqH4YRkgD8NBd0UojtE1
# XwYSBFUwDgYDVR0PAQH/BAQDAgGGMBIGA1UdEwEB/wQIMAYBAf8CAQAwEwYDVR0l
# BAwwCgYIKwYBBQUHAwgwEQYDVR0gBAowCDAGBgRVHSAAMFAGA1UdHwRJMEcwRaBD
# oEGGP2h0dHA6Ly9jcmwudXNlcnRydXN0LmNvbS9VU0VSVHJ1c3RSU0FDZXJ0aWZp
# Y2F0aW9uQXV0aG9yaXR5LmNybDB2BggrBgEFBQcBAQRqMGgwPwYIKwYBBQUHMAKG
# M2h0dHA6Ly9jcnQudXNlcnRydXN0LmNvbS9VU0VSVHJ1c3RSU0FBZGRUcnVzdENB
# LmNydDAlBggrBgEFBQcwAYYZaHR0cDovL29jc3AudXNlcnRydXN0LmNvbTANBgkq
# hkiG9w0BAQwFAAOCAgEAbVSBpTNdFuG1U4GRdd8DejILLSWEEbKw2yp9KgX1vDsn
# 9FqguUlZkClsYcu1UNviffmfAO9Aw63T4uRW+VhBz/FC5RB9/7B0H4/GXAn5M17q
# oBwmWFzztBEP1dXD4rzVWHi/SHbhRGdtj7BDEA+N5Pk4Yr8TAcWFo0zFzLJTMJWk
# 1vSWVgi4zVx/AZa+clJqO0I3fBZ4OZOTlJux3LJtQW1nzclvkD1/RXLBGyPWwlWE
# ZuSzxWYG9vPWS16toytCiiGS/qhvWiVwYoFzY16gu9jc10rTPa+DBjgSHSSHLeT8
# AtY+dwS8BDa153fLnC6NIxi5o8JHHfBd1qFzVwVomqfJN2Udvuq82EKDQwWli6YJ
# /9GhlKZOqj0J9QVst9JkWtgqIsJLnfE5XkzeSD2bNJaaCV+O/fexUpHOP4n2HKG1
# qXUfcb9bQ11lPVCBbqvw0NP8srMftpmWJvQ8eYtcZMzN7iea5aDADHKHwW5NWtMe
# 6vBE5jJvHOsXTpTDeGUgOw9Bqh/poUGd/rG4oGUqNODeqPk85sEwu8CgYyz8XBYA
# qNDEf+oRnR4GxqZtMl20OAkrSQeq/eww2vGnL8+3/frQo4TZJ577AWZ3uVYQ4SBu
# xq6x+ba6yDVdM3aO8XwgDCp3rrWiAoa6Ke60WgCxjKvj+QrJVF3UuWp0nr1Irpgw
# ggcHMIIE76ADAgECAhEAjHegAI/00bDGPZ86SIONazANBgkqhkiG9w0BAQwFADB9
# MQswCQYDVQQGEwJHQjEbMBkGA1UECBMSR3JlYXRlciBNYW5jaGVzdGVyMRAwDgYD
# VQQHEwdTYWxmb3JkMRgwFgYDVQQKEw9TZWN0aWdvIExpbWl0ZWQxJTAjBgNVBAMT
# HFNlY3RpZ28gUlNBIFRpbWUgU3RhbXBpbmcgQ0EwHhcNMjAxMDIzMDAwMDAwWhcN
# MzIwMTIyMjM1OTU5WjCBhDELMAkGA1UEBhMCR0IxGzAZBgNVBAgTEkdyZWF0ZXIg
# TWFuY2hlc3RlcjEQMA4GA1UEBxMHU2FsZm9yZDEYMBYGA1UEChMPU2VjdGlnbyBM
# aW1pdGVkMSwwKgYDVQQDDCNTZWN0aWdvIFJTQSBUaW1lIFN0YW1waW5nIFNpZ25l
# ciAjMjCCAiIwDQYJKoZIhvcNAQEBBQADggIPADCCAgoCggIBAJGHSyyLwfEeoJ7T
# B8YBylKwvnl5XQlmBi0vNX27wPsn2kJqWRslTOrvQNaafjLIaoF9tFw+VhCBNToi
# Noz7+CAph6x00BtivD9khwJf78WA7wYc3F5Ok4e4mt5MB06FzHDFDXvsw9njl+nL
# GdtWRWzuSyBsyT5s/fCb8Sj4kZmq/FrBmoIgOrfv59a4JUnCORuHgTnLw7c6zZ9Q
# BB8amaSAAk0dBahV021SgIPmbkilX8GJWGCK7/GszYdjGI50y4SHQWljgbz2H6p8
# 18FBzq2rdosggNQtlQeNx/ULFx6a5daZaVHHTqadKW/neZMNMmNTrszGKYogwWDG
# 8gIsxPnIIt/5J4Khg1HCvMmCGiGEspe81K9EHJaCIpUqhVSu8f0+SXR0/I6uP6Vy
# 9MNaAapQpYt2lRtm6+/a35Qu2RrrTCd9TAX3+CNdxFfIJgV6/IEjX1QJOCpi1arK
# 3+3PU6sf9kSc1ZlZxVZkW/eOUg9m/Jg/RAYTZG7p4RVgUKWx7M+46MkLvsWE990K
# ndq8KWw9Vu2/eGe2W8heFBy5r4Qtd6L3OZU3b05/HMY8BNYxxX7vPehRfnGtJHQb
# LNz5fKrvwnZJaGLVi/UD3759jg82dUZbk3bEg+6CviyuNxLxvFbD5K1Dw7dmll6U
# Mvqg9quJUPrOoPMIgRrRRKfM97gxAgMBAAGjggF4MIIBdDAfBgNVHSMEGDAWgBQa
# ofhhGSAPw0F3RSiO0TVfBhIEVTAdBgNVHQ4EFgQUaXU3e7udNUJOv1fTmtufAdGu
# 3tAwDgYDVR0PAQH/BAQDAgbAMAwGA1UdEwEB/wQCMAAwFgYDVR0lAQH/BAwwCgYI
# KwYBBQUHAwgwQAYDVR0gBDkwNzA1BgwrBgEEAbIxAQIBAwgwJTAjBggrBgEFBQcC
# ARYXaHR0cHM6Ly9zZWN0aWdvLmNvbS9DUFMwRAYDVR0fBD0wOzA5oDegNYYzaHR0
# cDovL2NybC5zZWN0aWdvLmNvbS9TZWN0aWdvUlNBVGltZVN0YW1waW5nQ0EuY3Js
# MHQGCCsGAQUFBwEBBGgwZjA/BggrBgEFBQcwAoYzaHR0cDovL2NydC5zZWN0aWdv
# LmNvbS9TZWN0aWdvUlNBVGltZVN0YW1waW5nQ0EuY3J0MCMGCCsGAQUFBzABhhdo
# dHRwOi8vb2NzcC5zZWN0aWdvLmNvbTANBgkqhkiG9w0BAQwFAAOCAgEASgN4kEIz
# 7Hsagwk2M5hVu51ABjBrRWrxlA4ZUP9bJV474TnEW7rplZA3N73f+2Ts5YK3lcxX
# VXBLTvSoh90ihaZXu7ghJ9SgKjGUigchnoq9pxr1AhXLRFCZjOw+ugN3poICkMIu
# k6m+ITR1Y7ngLQ/PATfLjaL6uFqarqF6nhOTGVWPCZAu3+qIFxbradbhJb1FCJeA
# 11QgKE/Ke7OzpdIAsGA0ZcTjxcOl5LqFqnpp23WkPnlomjaLQ6421GFyPA6FYg2g
# XnDbZC8Bx8GhxySUo7I8brJeotD6qNG4JRwW5sDVf2gaxGUpNSotiLzqrnTWgufA
# iLjhT3jwXMrAQFzCn9UyHCzaPKw29wZSmqNAMBewKRaZyaq3iEn36AslM7U/ba+f
# XwpW3xKxw+7OkXfoIBPpXCTH6kQLSuYThBxN6w21uIagMKeLoZ+0LMzAFiPJkeVC
# A0uAzuRN5ioBPsBehaAkoRdA1dvb55gQpPHqGRuAVPpHieiYgal1wA7f0GiUeaGg
# no62t0Jmy9nZay9N2N4+Mh4g5OycTUKNncczmYI3RNQmKSZAjngvue76L/Hxj/5Q
# uHjdFJbeHA5wsCqFarFsaOkq5BArbiH903ydN+QqBtbD8ddo408HeYEIE/6yZF7p
# sTzm0Hgjsgks4iZivzupl1HMx0QygbKvz98xggWDMIIFfwIBATCBkTB8MQswCQYD
# VQQGEwJHQjEbMBkGA1UECBMSR3JlYXRlciBNYW5jaGVzdGVyMRAwDgYDVQQHEwdT
# YWxmb3JkMRgwFgYDVQQKEw9TZWN0aWdvIExpbWl0ZWQxJDAiBgNVBAMTG1NlY3Rp
# Z28gUlNBIENvZGUgU2lnbmluZyBDQQIRAJ0ixooo0boZelXoKg5DAFQwCQYFKw4D
# AhoFAKB4MBgGCisGAQQBgjcCAQwxCjAIoAKAAKECgAAwGQYJKoZIhvcNAQkDMQwG
# CisGAQQBgjcCAQQwHAYKKwYBBAGCNwIBCzEOMAwGCisGAQQBgjcCARUwIwYJKoZI
# hvcNAQkEMRYEFAZM4lY86+RqRuCX7WO/DyedIoqhMA0GCSqGSIb3DQEBAQUABIIB
# ACq6BgtbMaBaNGoidFk1vq8Ixy1EeH4ptd0uQPmry5LnWAAVOiceXlPlw4CQILhx
# QvIO6ijXMVKECTKvIYTAgFXFudhQWJurE0MhbPT3c35lKBAiw6ZzIugpNEknu7al
# hnsRKfL49YK8ErceBoa8x53B+vqJqOSPpbDe2CcwUchjDD1H+E15i5lmhWU1HP/h
# kFZQdZPiF2LWSESoJgKTlbldvKKIsi3XVnBWC/SbGZRUOBVMFY1v53kvEkyIDLl9
# FoyNP7KBpQhJSG7ifv30DE0z6O5QeBWEn/LtxwZi9KQumlMgLD/ad0uIu5/kvEQ6
# fwT5W0kEX3q/c3DVC0bl0oihggNMMIIDSAYJKoZIhvcNAQkGMYIDOTCCAzUCAQEw
# gZIwfTELMAkGA1UEBhMCR0IxGzAZBgNVBAgTEkdyZWF0ZXIgTWFuY2hlc3RlcjEQ
# MA4GA1UEBxMHU2FsZm9yZDEYMBYGA1UEChMPU2VjdGlnbyBMaW1pdGVkMSUwIwYD
# VQQDExxTZWN0aWdvIFJTQSBUaW1lIFN0YW1waW5nIENBAhEAjHegAI/00bDGPZ86
# SIONazANBglghkgBZQMEAgIFAKB5MBgGCSqGSIb3DQEJAzELBgkqhkiG9w0BBwEw
# HAYJKoZIhvcNAQkFMQ8XDTIyMDMyNDA5MDg0OFowPwYJKoZIhvcNAQkEMTIEMLGl
# 1Z7IUTcXNaUKHpw+uTjLMh1aLHpVNK7fvr4Rygv/TQwXSo0NV1Ay6pwZzLIp1DAN
# BgkqhkiG9w0BAQEFAASCAgBKQcDRE+UX9gsKpD8pr34VZwtxpt2/WsfPIo7SUC1c
# bdAgQUGNijdOauQJ+tdg90Vl6m7KvCzxUHYhiSWr84ck3AxERboJ56Iu6tay07M0
# 9n0q7/aVChLZk+asr2JpW2Lfby9Ml6AE7snphwvJPSfRX4iqqzjsD+6Z9LtNkyD8
# bq/4TZjFWyooefgpZ0zKsxn7kiHkHBLY81bMg2a/l/67Pa+ll3ofsnBw7T7BDU6+
# SKjYulDuxFwy3+mhWPwrtCprZxJdwOEo5uX5plmlr0IbNAuc0D9zk7alDQpmqSVZ
# YPQ9hxTV9bpqIH5A2xu2Nl5pIt03zR3NHNfhNA9Zy0hGlmyL6XOpxO2smyD90s+Y
# iB/vHDlEEzoNU/Xvo4g1xmnONp0Snr2WgElUCFZ1d7e1205EdPQGd/pCW3YevT/X
# MBZiadNw/ECEgyR9wuKxhALqbrZMmae3AzlI5xrPLeW94BMKtI4APzYLsH1BivtG
# vVXk7BKM5FLYw59K324zPRhB/rAj81oG30ly45D9Lf2Jkm5Bd02V8HBoCVFds8tH
# 5DBI7lhPqmJqbFbLH4bKVCQpPxXAPcMeiSTEehhcbJ/owB1A7q+wglz8QyFzDr/i
# Xsm1102soJfbzQ7zqj9VRIpB+N5DZ5loBDW0tg+qxRJsrBHaEYPFiIWSVD3NjX1v
# KQ==
# SIG # End signature block
